# !usr/bin/env python
# -*- coding: utf-8 -*-
#
# Licensed under a 3-clause BSD license.
#
# @Author: Brian Cherinka
# @Date:   2017-09-13 16:05:56
# @Last modified by: José Sánchez-Gallego (gallegoj@uw.edu)
# @Last modified time: 2018-08-06 11:45:33

from __future__ import absolute_import, division, print_function

import copy
from astropy import units as u

from marvin.utils.datamodel.maskbit import get_maskbits

from .base import Bintype, Channel, DAPDataModel, Model, MultiChannelProperty, Property
from .base import spaxel as spaxel_unit
from .MPL5 import ALL, GAU_MILESHC, NRE, SPX, VOR10


HYB10 = Bintype('HYB10', description='Binning and stellar continuum fitting as VOR10, '
                                     'but emission lines are fitted per spaxel.')

# The two lines in the OII doublet is fitted independently for gaussian
# measurements. In that case oii_3727 and oii_3729 are populated. For summed
# flux measurements, the lines cannot be separated so oiid_3728 contains
# the summed flux. In that case, oii_3729 is null and only kept to maintain`
# the number of channels constant.
oiid_channel = Channel('oiid_3728', formats={'string': 'OIId 3728',
                       'latex': r'$\forb{O\,IId}\;\lambda\lambda 3728$'}, idx=0)
oii_channel = Channel('oii_3727', formats={'string': 'OII 3727',
                      'latex': r'$\forb{O\,II}\;\lambda 3727$'}, idx=0)

MPL6_emline_channels = [
    Channel('oii_3729', formats={'string': 'OII 3729',
                                 'latex': r'$\forb{O\,II}\;\lambda 3729$'}, idx=1),
    Channel('hthe_3798', formats={'string': 'H-theta 3798',
                                  'latex': r'H$\theta\;\lambda 3798$'}, idx=2),
    Channel('heta_3836', formats={'string': 'H-eta 3836',
                                  'latex': r'H$\eta\;\lambda 3836$'}, idx=3),
    Channel('neiii_3869', formats={'string': 'NeIII 3869',
                                   'latex': r'$\forb{Ne\,III}\;\lambda 3869$'}, idx=4),
    Channel('hzet_3890', formats={'string': 'H-zeta 3890',
                                  'latex': r'H$\zeta\;\lambda 3890$'}, idx=5),
    Channel('neiii_3968', formats={'string': 'NeIII 3968',
                                   'latex': r'$\forb{Ne\,III}\;\lambda 3968$'}, idx=6),
    Channel('heps_3971', formats={'string': 'H-epsilon 3971',
                                  'latex': r'H$\epsilon\;\lambda 3971$'}, idx=7),
    Channel('hdel_4102', formats={'string': 'H-delta 4102',
                                  'latex': r'H$\delta\;\lambda 4102$'}, idx=8),
    Channel('hgam_4341', formats={'string': 'H-gamma 4341',
                                  'latex': r'H$\gamma\;\lambda 4341$'}, idx=9),
    Channel('heii_4687', formats={'string': 'HeII 4681',
                                  'latex': r'He\,II$\;\lambda 4687$'}, idx=10),
    Channel('hb_4862', formats={'string': 'H-beta 4862',
                                'latex': r'H$\beta\;\lambda 4862$'}, idx=11),
    Channel('oiii_4960', formats={'string': 'OIII 4960',
                                  'latex': r'$\forb{O\,III}\;\lambda 4960$'}, idx=12),
    Channel('oiii_5008', formats={'string': 'OIII 5008',
                                  'latex': r'$\forb{O\,III}\;\lambda 5008$'}, idx=13),
    Channel('hei_5877', formats={'string': 'HeI 5877',
                                 'latex': r'He\,I$\;\lambda 5877$'}, idx=14),
    Channel('oi_6302', formats={'string': 'OI 6302',
                                'latex': r'$\forb{O\,I}\;\lambda 6302$'}, idx=15),
    Channel('oi_6365', formats={'string': 'OI 6365',
                                'latex': r'$\forb{O\,I}\;\lambda 6365$'}, idx=16),
    Channel('nii_6549', formats={'string': 'NII 6549',
                                 'latex': r'$\forb{N\,II}\;\lambda 6549$'}, idx=17),
    Channel('ha_6564', formats={'string': 'H-alpha 6564',
                                'latex': r'H$\alpha\;\lambda 6564$'}, idx=18),
    Channel('nii_6585', formats={'string': 'NII 6585',
                                 'latex': r'$\forb{N\,II}\;\lambda 6585$'}, idx=19),
    Channel('sii_6718', formats={'string': 'SII 6718',
                                 'latex': r'$\forb{S\,II}\;\lambda 6718$'}, idx=20),
    Channel('sii_6732', formats={'string': 'SII 6732',
                                 'latex': r'$\forb{S\,II}\;\lambda 6732$'}, idx=21)
]


MPL6_specindex_channels = [
    Channel('cn1', formats={'string': 'CN1'}, unit=u.mag, idx=0),
    Channel('cn2', formats={'string': 'CN2'}, unit=u.mag, idx=1),
    Channel('ca4227', formats={'string': 'Ca 4227',
                               'latex': r'Ca\,\lambda 4227'}, unit=u.Angstrom, idx=2),
    Channel('g4300', formats={'string': 'G4300',
                              'latex': r'G\,\lambda 4300'}, unit=u.Angstrom, idx=3),
    Channel('fe4383', formats={'string': 'Fe 4383',
                               'latex': r'Fe\,\lambda 4383'}, unit=u.Angstrom, idx=4),
    Channel('ca4455', formats={'string': 'Ca 4455',
                               'latex': r'Ca\,\lambda 4455'}, unit=u.Angstrom, idx=5),
    Channel('fe4531', formats={'string': 'Fe 4531',
                               'latex': r'Fe\,\lambda 4531'}, unit=u.Angstrom, idx=6),
    Channel('c24668', formats={'string': 'C24668',
                               'latex': r'C2\,\lambda 4668'}, unit=u.Angstrom, idx=7),
    Channel('hb', formats={'string': 'Hb',
                           'latex': r'H\beta'}, unit=u.Angstrom, idx=8),
    Channel('fe5015', formats={'string': 'Fe 5015',
                               'latex': r'Fe\,\lambda 5015'}, unit=u.Angstrom, idx=9),
    Channel('mg1', formats={'string': 'Mg1'}, unit=u.mag, idx=10),
    Channel('mg2', formats={'string': 'Mg2'}, unit=u.mag, idx=11),
    Channel('mgb', formats={'string': 'Mgb'}, unit=u.Angstrom, idx=12),
    Channel('fe5270', formats={'string': 'Fe 5270',
                               'latex': r'Fe\,\lambda 5270'}, unit=u.Angstrom, idx=13),
    Channel('fe5335', formats={'string': 'Fe 5335',
                               'latex': r'Fe\,\lambda 5335'}, unit=u.Angstrom, idx=14),
    Channel('fe5406', formats={'string': 'Fe 5406',
                               'latex': r'Fe\,\lambda 5406'}, unit=u.Angstrom, idx=15),
    Channel('fe5709', formats={'string': 'Fe 5709',
                               'latex': r'Fe\,\lambda 5709'}, unit=u.Angstrom, idx=16),
    Channel('fe5782', formats={'string': 'Fe 5782',
                               'latex': r'Fe\,\lambda 5782'}, unit=u.Angstrom, idx=17),
    Channel('nad', formats={'string': 'NaD'}, unit=u.Angstrom, idx=18),
    Channel('tio1', formats={'string': 'TiO1'}, unit=u.mag, idx=19),
    Channel('tio2', formats={'string': 'TiO2'}, unit=u.mag, idx=20),
    Channel('hdeltaa', formats={'string': 'HDeltaA',
                                'latex': r'H\delta\,A'}, unit=u.Angstrom, idx=21),
    Channel('hgammaa', formats={'string': 'HGammaA',
                                'latex': r'H\gamma\,F'}, unit=u.Angstrom, idx=22),
    Channel('hdeltaf', formats={'string': 'HDeltaA',
                                'latex': r'H\delta\,F'}, unit=u.Angstrom, idx=23),
    Channel('hgammaf', formats={'string': 'HGammaF',
                                'latex': r'H\gamma\,F'}, unit=u.Angstrom, idx=24),
    Channel('cahk', formats={'string': 'CaHK'}, unit=u.Angstrom, idx=25),
    Channel('caii1', formats={'string': 'CaII1'}, unit=u.Angstrom, idx=26),
    Channel('caii2', formats={'string': 'CaII2'}, unit=u.Angstrom, idx=27),
    Channel('caii3', formats={'string': 'CaII3'}, unit=u.Angstrom, idx=28),
    Channel('pa17', formats={'string': 'Pa17'}, unit=u.Angstrom, idx=29),
    Channel('pa14', formats={'string': 'Pa14'}, unit=u.Angstrom, idx=30),
    Channel('pa12', formats={'string': 'Pa12'}, unit=u.Angstrom, idx=31),
    Channel('mgicvd', formats={'string': 'MgICvD'}, unit=u.Angstrom, idx=32),
    Channel('naicvd', formats={'string': 'NaICvD'}, unit=u.Angstrom, idx=33),
    Channel('mgiir', formats={'string': 'MgIIR'}, unit=u.Angstrom, idx=34),
    Channel('fehcvd', formats={'string': 'FeHCvD'}, unit=u.Angstrom, idx=35),
    Channel('nai', formats={'string': 'NaI'}, unit=u.Angstrom, idx=36),
    Channel('btio', formats={'string': 'bTiO'}, unit=u.mag, idx=37),
    Channel('atio', formats={'string': 'aTiO'}, unit=u.mag, idx=38),
    Channel('cah1', formats={'string': 'CaH1'}, unit=u.mag, idx=39),
    Channel('cah2', formats={'string': 'CaH2'}, unit=u.mag, idx=40),
    Channel('naisdss', formats={'string': 'NaISDSS'}, unit=u.Angstrom, idx=41),
    Channel('tio2sdss', formats={'string': 'TiO2SDSS'}, unit=u.Angstrom, idx=42),
    Channel('d4000', formats={'string': 'D4000'}, unit=u.dimensionless_unscaled, idx=43),
    Channel('dn4000', formats={'string': 'Dn4000'}, unit=u.dimensionless_unscaled, idx=44),
    Channel('tiocvd', formats={'string': 'TiOCvD'}, unit=u.dimensionless_unscaled, idx=45)
]


MPL6_binid_channels = [
    Channel('binned_spectra', formats={'string': 'Binned spectra'},
            unit=u.dimensionless_unscaled, idx=0),
    Channel('stellar_continua', formats={'string': 'Stellar continua'},
            unit=u.dimensionless_unscaled, idx=1),
    Channel('em_line_moments', formats={'string': 'Emission line moments'},
            unit=u.dimensionless_unscaled, idx=2),
    Channel('em_line_models', formats={'string': 'Emission line models'},
            unit=u.dimensionless_unscaled, idx=3),
    Channel('spectral_indices', formats={'string': 'Spectral indices'},
            unit=u.dimensionless_unscaled, idx=4)]


binid_properties = MultiChannelProperty('binid', ivar=False, mask=False,
                                        channels=MPL6_binid_channels,
                                        description='Numerical ID for spatial bins.')


MPL6_maps = [
    MultiChannelProperty('spx_skycoo', ivar=False, mask=False,
                         channels=[Channel('on_sky_x', formats={'string': 'On-sky X'}, idx=0),
                                   Channel('on_sky_y', formats={'string': 'On-sky Y'}, idx=1)],
                         unit=u.arcsec,
                         formats={'string': 'Sky coordinates'},
                         description='Offsets of each spaxel from the galaxy center.'),
    MultiChannelProperty('spx_ellcoo', ivar=False, mask=False,
                         channels=[Channel('elliptical_radius',
                                           formats={'string': 'Elliptical radius'},
                                           idx=0, unit=u.arcsec),
                                   Channel('r_re',
                                           formats={'string': 'R/Reff'},
                                           idx=1),
                                   Channel('elliptical_azimuth',
                                           formats={'string': 'Elliptical azimuth'},
                                           idx=2, unit=u.deg)],
                         formats={'string': 'Elliptical coordinates'},
                         description='Elliptical polar coordinates of each spaxel from '
                                     'the galaxy center.'),
    Property('spx_mflux', ivar=True, mask=False,
             unit=u.erg / u.s / (u.cm ** 2) / spaxel_unit, scale=1e-17,
             formats={'string': 'r-band mean flux'},
             description='Mean flux in r-band (5600.1-6750.0 ang).'),
    Property('spx_snr', ivar=False, mask=False,
             formats={'string': 'r-band SNR'},
             description='r-band signal-to-noise ratio per pixel.'),
    binid_properties,
    MultiChannelProperty('bin_lwskycoo', ivar=False, mask=False,
                         channels=[Channel('lum_weighted_on_sky_x',
                                           formats={'string': 'Light-weighted offset X'},
                                           idx=0, unit=u.arcsec),
                                   Channel('lum_weighted_on_sky_y',
                                           formats={'string': 'Light-weighted offset Y'},
                                           idx=1, unit=u.arcsec)],
                         description='Light-weighted offset of each bin from the galaxy center.'),
    MultiChannelProperty('bin_lwellcoo', ivar=False, mask=False,
                         channels=[Channel('lum_weighted_elliptical_radius',
                                           formats={'string': 'Light-weighted radial offset'},
                                           idx=0, unit=u.arcsec),
                                   Channel('r_re',
                                           formats={'string': 'R/REff'},
                                           idx=1),
                                   Channel('lum_weighted_elliptical_azimuth',
                                           formats={'string': 'Light-weighted azimuthal offset'},
                                           idx=2, unit=u.deg)],
                         description='Light-weighted elliptical polar coordinates of each bin '
                                     'from the galaxy center.'),
    Property('bin_area', ivar=False, mask=False,
             unit=u.arcsec ** 2,
             formats={'string': 'Bin area'},
             description='Area of each bin.'),
    Property('bin_farea', ivar=False, mask=False,
             formats={'string': 'Bin fractional area'},
             description='Fractional area that the bin covers for the expected bin '
                         'shape (only relevant for radial binning).'),
    Property('bin_mflux', ivar=True, mask=True,
             unit=u.erg / u.s / (u.cm ** 2) / spaxel_unit, scale=1e-17,
             formats={'string': 'r-band binned spectra mean flux'},
             description='Mean flux in the r-band for the binned spectra.'),
    Property('bin_snr', ivar=False, mask=False,
             formats={'string': 'Bin SNR'},
             description='r-band signal-to-noise ratio per pixel in the binned spectra.'),
    Property('stellar_vel', ivar=True, mask=True,
             unit=u.km / u.s,
             formats={'string': 'Stellar velocity'},
             description='Stellar velocity relative to NSA redshift.'),
    Property('stellar_sigma', ivar=True, mask=True,
             unit=u.km / u.s,
             formats={'string': 'Stellar velocity dispersion', 'latex': r'Stellar $\sigma$'},
             description='Stellar velocity dispersion (must be corrected using '
                         'STELLAR_SIGMACORR)'),
    Property('stellar_sigmacorr', ivar=False, mask=False,
             unit=u.km / u.s,
             formats={'string': 'Stellar sigma correction',
                      'latex': r'Stellar $\sigma$ correction'},
             description='Quadrature correction for STELLAR_SIGMA to obtain the '
                         'astrophysical velocity dispersion.)'),
    MultiChannelProperty('stellar_cont_fresid', ivar=False, mask=False,
                         channels=[Channel('68th_percentile',
                                           formats={'string': '68th percentile',
                                                    'latex': r'68^{th} percentile'}, idx=0),
                                   Channel('99th_percentile',
                                           formats={'string': '99th percentile',
                                                    'latex': r'99^{th} percentile'}, idx=1)],
                         formats={'string': 'Fractional residual growth'},
                         description='68%% and 99%% growth of the fractional residuals between '
                                     'the model and data.'),
    Property('stellar_cont_rchi2', ivar=False, mask=False,
             formats={'string': 'Stellar continuum reduced chi-square',
                      'latex': r'Stellar\ continuum\ reduced\ \chi^2'},
             description='Reduced chi-square of the stellar continuum fit.'),
    MultiChannelProperty('emline_sflux', ivar=True, mask=True,
                         channels=[oiid_channel] + MPL6_emline_channels,
                         formats={'string': 'Emission line summed flux'},
                         unit=u.erg / u.s / (u.cm ** 2) / spaxel_unit, scale=1e-17,
                         binid=binid_properties[3],
                         description='Non-parametric summed flux for emission lines.'),
    MultiChannelProperty('emline_sew', ivar=True, mask=True,
                         channels=[oiid_channel] + MPL6_emline_channels,
                         formats={'string': 'Emission line EW'},
                         unit=u.Angstrom,
                         binid=binid_properties[3],
                         description='Emission line non-parametric equivalent '
                                     'widths measurements.'),
    MultiChannelProperty('emline_gflux', ivar=True, mask=True,
                         channels=[oii_channel] + MPL6_emline_channels,
                         formats={'string': 'Emission line Gaussian flux'},
                         unit=u.erg / u.s / (u.cm ** 2) / spaxel_unit, scale=1e-17,
                         binid=binid_properties[3],
                         description='Gaussian profile integrated flux for emission lines.'),
    MultiChannelProperty('emline_gvel', ivar=True, mask=True,
                         channels=[oii_channel] + MPL6_emline_channels,
                         formats={'string': 'Emission line Gaussian velocity'},
                         unit=u.km / u.s,
                         binid=binid_properties[3],
                         description='Gaussian profile velocity for emission lines.'),
    MultiChannelProperty('emline_gew', ivar=True, mask=True,
                         channels=[oii_channel] + MPL6_emline_channels,
                         formats={'string': 'Emission line Gaussian EW'},
                         unit=u.Angstrom,
                         binid=binid_properties[3],
                         description='Gaussian-fitted equivalent widths measurements '
                                     '(based on EMLINE_GFLUX).'),
    MultiChannelProperty('emline_gsigma', ivar=True, mask=True,
                         channels=[oii_channel] + MPL6_emline_channels,
                         formats={'string': 'Emission line Gaussian sigma',
                                  'latex': r'Emission line Gaussian $\sigma$'},
                         unit=u.km / u.s,
                         binid=binid_properties[3],
                         description='Gaussian profile velocity dispersion for emission lines; '
                                     'must be corrected using EMLINE_INSTSIGMA.'),
    MultiChannelProperty('emline_instsigma', ivar=False, mask=False,
                         channels=[oii_channel] + MPL6_emline_channels,
                         formats={'string': 'Emission line instrumental sigma',
                                  'latex': r'Emission line instrumental $\sigma$'},
                         unit=u.km / u.s,
                         binid=binid_properties[3],
                         description='Instrumental dispersion at the fitted line center.'),
    MultiChannelProperty('emline_tplsigma', ivar=False, mask=False,
                         channels=[oii_channel] + MPL6_emline_channels,
                         formats={'string': 'Emission line template instrumental sigma',
                                  'latex': r'Emission line template instrumental $\sigma$'},
                         unit=u.km / u.s,
                         binid=binid_properties[3],
                         description='The dispersion of each emission line used in '
                                     'the template spectra'),
    MultiChannelProperty('specindex', ivar=True, mask=True,
                         channels=MPL6_specindex_channels,
                         formats={'string': 'Spectral index'},
                         description='Measurements of spectral indices.'),
    MultiChannelProperty('specindex_corr', ivar=False, mask=False,
                         channels=MPL6_specindex_channels,
                         formats={'string': 'Spectral index sigma correction',
                                  'latex': r'Spectral index $\sigma$ correction'},
                         description='Velocity dispersion corrections for the '
                                     'spectral index measurements '
                                     '(can be ignored for D4000, Dn4000).')
]


MPL6_models = [
    Model('binned_flux', 'FLUX', 'WAVE', extension_ivar='IVAR',
          extension_mask='MASK', unit=u.erg / u.s / (u.cm ** 2) / u.Angstrom / spaxel_unit,
          scale=1e-17, formats={'string': 'Binned flux'},
          description='Flux of the binned spectra',
          binid=binid_properties[0]),
    Model('full_fit', 'MODEL', 'WAVE', extension_ivar=None,
          extension_mask='MASK', unit=u.erg / u.s / (u.cm ** 2) / u.Angstrom / spaxel_unit,
          scale=1e-17, formats={'string': 'Best fitting model'},
          description='The best fitting model spectra (sum of the fitted '
                      'continuum and emission-line models)',
          binid=binid_properties[0]),
    Model('emline_fit', 'EMLINE', 'WAVE', extension_ivar=None,
          extension_mask='EMLINE_MASK',
          unit=u.erg / u.s / (u.cm ** 2) / u.Angstrom / spaxel_unit,
          scale=1e-17, formats={'string': 'Emission line model spectrum'},
          description='The model spectrum with only the emission lines.',
          binid=binid_properties[3]),
    Model('emline_base_fit', 'EMLINE_BASE', 'WAVE', extension_ivar=None,
          extension_mask='EMLINE_MASK',
          unit=u.erg / u.s / (u.cm ** 2) / u.Angstrom / spaxel_unit,
          scale=1e-17, formats={'string': 'Emission line baseline fit'},
          description='The model of the constant baseline fitted beneath the '
                      'emission lines.',
          binid=binid_properties[3])
]


# MPL-6 DapDataModel goes here
MPL6 = DAPDataModel('2.1.3', aliases=['MPL-6', 'MPL6'],
                    bintypes=[SPX, HYB10, VOR10, ALL, NRE],
                    db_only=[SPX, HYB10],
                    templates=[GAU_MILESHC],
                    properties=MPL6_maps,
                    models=MPL6_models,
                    bitmasks=get_maskbits('MPL-6'),
                    default_bintype='SPX',
                    default_template='GAU-MILESHC',
                    property_table='SpaxelProp6',
                    default_binid=copy.deepcopy(binid_properties[0]),
                    default_mapmask=['NOCOV', 'UNRELIABLE', 'DONOTUSE'],
                    qual_flag='DAPQUAL')



# !usr/bin/env python
# -*- coding: utf-8 -*-
#
# Licensed under a 3-clause BSD license.
#
# @Author: Brian Cherinka
# @Date:   2018-10-11 17:51:43
# @Last modified by:   Brian Cherinka
# @Last Modified time: 2018-11-29 17:23:15

from __future__ import print_function, division, absolute_import

from astropy.io import fits
import marvin.tools
from marvin import log

from .base import VACMixIn


class GEMAVAC(VACMixIn):
    """Provides access to the MaNGA-GEMA VAC.

    VAC name: GEMA

    URL: https://www.sdss.org/dr17/data_access/value-added-catalogs/?vac_id=gema-vac:-galaxy-environment-for-manga-value-added-catalog

    Description: The GEMA VAC contains many different quantifications of the local and the large-scale environments for MaNGA galaxies. Please visit the DATAMODEL at https://data.sdss.org/datamodel/files/MANGA_GEMA/GEMA_VER to see the description of each table composing the catalogue.

    Authors: Maria Argudo-Fernandez, Daniel Goddard, Daniel Thomas, Zheng Zheng, Lihwai Lin, Ting Xiao, Fangting Yuan, Jianhui Lian, et al

    """

    # Required parameters
    name = 'gema'
    description = 'Returns GEMA table data'
    version = {'MPL-7': '1.0.1', 'DR15': '1.0.1', 'DR16': '1.0.1', 'DR17': '2.0.2', 
               'MPL-11': '2.0.2'}
    display_name = 'GEMA'
    url = 'https://www.sdss.org/dr17/data_access/value-added-catalogs/?vac_id=gema-vac:-galaxy-environment-for-manga-value-added-catalog'

    # optional Marvin Tools to attach your vac to
    include = (marvin.tools.cube.Cube, marvin.tools.maps.Maps, marvin.tools.modelcube.ModelCube)

    # Required method
    def set_summary_file(self, release):
        ''' Sets the path to the GEMA summary file '''

        # define the variables to build a unique path to your VAC file
        self.path_params = {"ver": self.version[release]}

        # get_path returns False if the files do not exist locally
        self.summary_file = self.get_path('mangagema', path_params=self.path_params)

    # Required method
    def get_target(self, parent_object):
        ''' Accesses VAC data for a specific target from a Marvin Tool object '''

        # Create a dictionary for GEMA VAC data
        gemadata = {
            "completeness": "",
            "LSS_1_all": "",
            "LSS_1_002": "",
            "LSS_1_006": "",
            "LSS_1_010": "",
            "LSS_1_015": "",
            "LSS_5_all": "",
            "LSS_5_002": "",
            "LSS_5_006": "",
            "LSS_5_010": "",
            "LSS_5_015": "",
            "pairs": "",
            "groups": "",
            "overdensity": "",
            "LSS_tensor": "",
        }

        # get any parameters you need from the parent object
        mangaid = parent_object.mangaid

        # download the vac from the SAS if it does not already exist locally
        if not self.file_exists(self.summary_file):
            self.summary_file = self.download_vac('mangagema', path_params=self.path_params)

        # opening tables in VAC file
        gemafile = fits.open(self.summary_file)

        # Return selected line(s) for a mangaid galaxy
        # LSS parameters
        indata = mangaid in gemafile[1].data['mangaid']
        if not indata:
            log.warning("No LSS data exists for {0}".format(mangaid))
        else:
            log.info("LSS data exists for {0}".format(mangaid))
            log.warning(
                "Warning: Do not use LSS parameters defined in volume with z < z_manga_galaxy"
            )
            indexLSS = gemafile[1].data['mangaid'] == mangaid
            gemadata["completeness"] = gemafile[1].data[indexLSS]
            gemadata["LSS_1_all"] = gemafile[2].data[indexLSS]
            gemadata["LSS_1_002"] = gemafile[3].data[indexLSS]
            gemadata["LSS_1_006"] = gemafile[4].data[indexLSS]
            gemadata["LSS_1_010"] = gemafile[5].data[indexLSS]
            gemadata["LSS_1_015"] = gemafile[6].data[indexLSS]
            gemadata["LSS_5_all"] = gemafile[7].data[indexLSS]
            gemadata["LSS_5_002"] = gemafile[8].data[indexLSS]
            gemadata["LSS_5_006"] = gemafile[9].data[indexLSS]
            gemadata["LSS_5_010"] = gemafile[10].data[indexLSS]
            gemadata["LSS_5_015"] = gemafile[11].data[indexLSS]

        # Local environment: close pair galaxies
        indata = mangaid in gemafile[12].data['mangaid']
        if not indata:
            log.warning("No pair data exists for {0}".format(mangaid))
        else:
            log.info("Pair data exists for {0}".format(mangaid))
            indexpair = gemafile[12].data['mangaid'] == mangaid
            gemadata["pairs"] = gemafile[12].data[indexpair]

        # Local environment: group galaxies
        indata = mangaid in gemafile[13].data['mangaid']
        if not indata:
            log.warning("No group data exists for {0}".format(mangaid))
        else:
            log.info("Group data exists for {0}".format(mangaid))
            indexgroup = gemafile[13].data['mangaid'] == mangaid
            gemadata["groups"] = gemafile[13].data[indexgroup]

        # LSS overdensity-corrected local density
        indata = mangaid in gemafile[14].data['mangaid']
        if not indata:
            log.warning("No overdensity data exists for {0}".format(mangaid))
        else:
            log.info("Overdensity data exists for {0}".format(mangaid))
            indexover = gemafile[14].data['mangaid'] == mangaid
            gemadata["overdensity"] = gemafile[14].data[indexover]

        # LSS structure
        indata = mangaid in gemafile[15].data['mangaid']
        if not indata:
            log.warning("No structure data exists for {0}".format(mangaid))
        else:
            log.info("Structure data exists for {0}".format(mangaid))
            indextensor = gemafile[15].data['mangaid'] == mangaid
            gemadata["LSS_tensor"] = gemafile[15].data[indextensor]

        gema_data = gemadata

        # closing the FITS file
        gemafile.close()

        return gema_data



# !usr/bin/env python
# -*- coding: utf-8 -*-
#
# Licensed under a 3-clause BSD license.
#
# @Author: Brian Cherinka
# @Date:   2018-08-04 20:09:38
# @Last modified by:   Brian Cherinka
# @Last modified time: 2018-11-22 13:11:86

from __future__ import print_function, division, absolute_import, unicode_literals

import os
import re
import warnings
import datetime
from collections import Counter, defaultdict
from functools import wraps
from operator import eq, ge, gt, le, lt, ne

import numpy as np
import six
from marvin import config, log
from marvin.api.api import Interaction
from marvin.core import marvin_pickle
from marvin.core.exceptions import MarvinError, MarvinUserWarning
from marvin.tools.results import Results, remote_mode_only
from marvin.utils.general import temp_setattr, getKeywordArgs
from marvin.utils.datamodel.query import datamodel
from marvin.utils.datamodel.query.base import query_params

if config.db:
    from marvin import marvindb
    from marvin.utils.general.structs import string_folding_wrapper
    from sqlalchemy import bindparam, func, and_
    from sqlalchemy.dialects import postgresql
    from sqlalchemy.orm import aliased
    from sqlalchemy.sql.expression import desc
    from sqlalchemy_boolean_search import (BooleanSearchException, parse_boolean_search)

try:
    import cPickle as pickle
except:
    import pickle


__all__ = ['Query', 'doQuery']

opdict = {'<=': le, '>=': ge, '>': gt, '<': lt, '!=': ne, '=': eq, '==': eq}


def doQuery(**kwargs):
    """Convenience function for building a Query and retrieving the Results.

    Parameters:
        N/A:
            See the :class:`~marvin.tools.query.Query` class for a list
            of inputs.

    Returns:
        query, results:
            A tuple containing the built
            :class:`~marvin.tools.query.Query` instance, and the
            :class:`~marvin.tools.results.Results` instance.
    """
    # pop some pagination keywords
    start = kwargs.pop('start', None)
    end = kwargs.pop('end', None)
    query_type = kwargs.pop('query_type', None)
    # get Query keyword arguments and check input kwargs
    qwargs = getKeywordArgs(Query)
    good_kwargs = {k: v for k, v in kwargs.items() if k in qwargs}
    # run the query
    q = Query(**good_kwargs)
    try:
        res = q.run(start=start, end=end, query_type=query_type)
    except TypeError as e:
        warnings.warn('Cannot run, query object is None: {0}.'.format(e), MarvinUserWarning)
        res = None

    return q, res


def update_config(f):
    """Decorator that updates query object with new config drpver and dapver versions."""

    @wraps(f)
    def wrapper(self, *args, **kwargs):
        if self.query and self.data_origin == 'db':
            self.query = self.query.params({'drpver': self._drpver, 'dapver': self._dapver})
        return f(self, *args, **kwargs)
    return wrapper


def tree():
    return defaultdict(tree)


class Query(object):
    ''' A class to perform queries on the MaNGA dataset.

    This class is the main way of performing a query.  A query works by minimally
    specifying a string filter a string filter condition in a natural language SQL format,
    as well as, a list of desired parameters to return.

    Query will use a local database if it finds on.  Otherwise a remote query uses
    the API to run a query on the Utah Server and return the results.

    The Query returns a list of tupled parameters and passed them into the
    Marvin Results object.  The parameters are a combination of user-defined
    return parameters, parameters used in the filter condition, and a set of pre-defined
    default parameters.  The object plateifu or mangaid is always returned by default.
    For queries involving DAP properties, the bintype, template, and spaxel x and y are
    also returned by default.

    Parameters:
        search_filter (str):
            A (natural language) string containing the filter conditions
            in the query.
        return_params (list):
            A list of string parameter names desired to be returned in the query
        return_type ({'cube', 'spaxel', 'maps', 'rss', and 'modelcube'}):
            The requested Marvin Tool object that the results are converted into.
        targets (list):
            A list of manga_target flags to filter on
        quality (list):
            A list of quality flags to filter on
        mode ({'local', 'remote', 'auto'}):
            The load mode to use. See :doc:`Mode secision tree</mode_decision>`.
        return_all (bool):
            If True, attempts to return the entire set of results. Default is False.
        default_params (list):
            Optionally specify additional parameters as defaults
        sort (str):
            The parameter name to sort the query on
        order ({'asc', 'desc'}):
            The sort order.  Can be either ascending or descending.
        limit (int):
            The number limit on the number of returned results
        count_threshold (int):
            The threshold number to begin paginating results.  Default is 1000.
        nexus (str):
            The name of the database table to use as the nexus point for building
            the join table tree.  Can only be set in local mode.
        caching (bool):
            If True, turns on the dogpile memcache caching of results. Default is True.
        verbose (bool):
            If True, turns on verbosity.

    '''

    def __init__(self, search_filter=None, return_params=None, return_type=None, targets=None,
                 quality=None, mode=None, return_all=False, default_params=None, nexus='cube',
                 sort='mangaid', order='asc', caching=True, limit=100, count_threshold=1000,
                 verbose=False, release=None):

        # basic parameters
        self.release = release or config.release
        self._drpver, self._dapver = config.lookUpVersions(release=self.release)
        self.mode = mode if mode is not None else config.mode
        self.data_origin = None

        # main parameters
        self.search_filter = search_filter
        self.return_params = return_params
        self.return_type = return_type
        self.default_params = default_params
        self.targets = targets
        self.quality = quality
        self.filter_params = {}
        self.params = []
        self._nexus = nexus
        self._results = None
        self.datamodel = datamodel[self.release]

        # optional parameters
        self.return_all = return_all
        self.sort = sort
        self.order = order
        self._caching = caching
        self.count_threshold = count_threshold
        self.limit = limit
        self.verbose = verbose

        # add db specific parameters
        if config.db:
            self._marvinform = self.datamodel._marvinform
            self.session = marvindb.session
            self._modelgraph = marvindb.modelgraph

        # timings
        self._run_time = None
        self._final_time = None

        # define the Query MMA
        self._set_mma()

        # initialize a query
        if self.data_origin == 'file':
            raise MarvinError('Cannot currently query a file')
        elif self.data_origin == 'db':
            self._init_local_query()
        elif self.data_origin == 'api':
            self._init_remote_query()

    def __repr__(self):
        return ('Marvin Query(filter={0}, mode={1}, data_origin={2})'
                .format(repr(self.search_filter), repr(self.mode), repr(self.data_origin)))

    @property
    def nexus(self):
        return self._nexus

    @nexus.setter
    def nexus(self, value):
        if not self.data_origin == 'db':
            raise MarvinError('Can only set nexus point with a local db origin')
        self._nexus = value

    def _set_mma(self):
        ''' Sets up the Query MMA system '''

        if self.mode == 'local':
            self._do_local()
        if self.mode == 'remote':
            self._do_remote()
        if self.mode == 'auto':
            try:
                self._do_local()
            except MarvinError as e:
                log.debug('local mode failed. Trying remote now.')
                self._do_remote()

        # Sanity check to make sure data_origin has been properly set.
        assert self.data_origin in ['file', 'db', 'api'], 'data_origin is not properly set.'

    def _do_local(self):
        ''' Sets up to perform queries locally. '''

        if not config.db:
            warnings.warn('No local database found. Cannot perform queries.', MarvinUserWarning)
            raise MarvinError('No local database found.  Query cannot be run in local mode')
        else:
            self.mode = 'local'
            self.data_origin = 'db'

    def _do_remote(self):
        ''' Sets up to perform queries remotely. '''

        if not config.urlmap:
            raise MarvinError('No URL Map found.  Cannot make remote query calls!')
        else:
            self.mode = 'remote'
            self.data_origin = 'api'

    def _init_local_query(self):
        ''' Initialize a local database query '''

        # set default parameters
        self._set_defaultparams()

        # get user-defined input parameters
        self._set_return_params()

        # setup the search filter
        self._set_filter()

        # build the query
        self._build_query()

    def _init_remote_query(self):
        ''' Initialize a remote API query '''

        # set up the parameters
        returns = ','.join(self.return_params) if self.return_params else None
        defaults = ','.join(self.default_params) if self.default_params else None
        targets = ','.join(self.targets) if self.targets else None
        quality = ','.join(self.quality) if self.quality else None
        search = self.search_filter if self.search_filter else ''

        self._remote_params = {'searchfilter': search,
                               'returnparams': returns,
                               'returntype': self.return_type,
                               'defaults': defaults,
                               'targets': targets,
                               'quality': quality,
                               'release': self.release,
                               'sort': self.sort,
                               'order': self.order,
                               'limit': self.limit,
                               'return_all': self.return_all,
                               'caching': self._caching}

    def run(self, start=None, end=None, query_type=None):
        ''' Runs a Query

        Runs a query either locally or remotely.

        Parameters:
            start (int):
                A starting index when slicing the query
            end (int):
                An ending index when slicing the query
            query_type (str):
                The type of SQLAlchemy to submit. Can be "raw", "core", "orm"

        Returns:
            An instance of the :class:`~marvin.tools.query.results.Results`
            class containing the results of your Query.

        Example:
            >>> # filter of "NSA redshift less than 0.1 and stellar mass > 1.e10"
            >>> searchfilter = 'nsa.z < 0.1 and nsa.elpetro_mass > 1.e10'
            >>> returnparams = ['cube.ra', 'cube.dec']
            >>> q = Query(search_filter=searchfilter, return_params=returnparams)
            >>> results = q.run()

        '''

        if self.data_origin == 'api':
            results = self._run_remote(start=start, end=end, query_type=query_type)
        elif self.data_origin == 'db':
            results = self._run_local(start=start, end=end, query_type=query_type)

        return results

    def _run_remote(self, start=None, end=None, query_type=None):
        ''' Run a remote Query

        Runs a query remotely.  Creates a dictionary of all input parameters and
        performs the appropriate API call.  On return, converts the JSON results
        into a Marvin Results object.

        Parameters:
            start (int):
                A starting index when slicing the query
            end (int):
                An ending index when slicing the query
            query_type (str):
                The type of SQLAlchemy to submit. Can be "raw", "core", "orm"

        Returns:
            An instance of the :class:`~marvin.tools.query.results.Results`
            class containing the results of your Query.

        '''

        if self.return_all:
            warnings.warn('Warning: Attempting to return all results. This may take a while or crash.')

        # Get the query route
        url = config.urlmap['api']['querycubes']['url']

        # Update the remote params
        self._remote_params.update({'start': start, 'end': end, 'query_type': query_type})

        # set the start time of query
        starttime = datetime.datetime.now()

        # Request the query
        try:
            ii = Interaction(route=url, params=self._remote_params, stream=True, datastream=self.return_all)
        except Exception as e:
            raise MarvinError('API Query call failed: {0}'.format(e))
        else:
            # retrive and set some parameters
            remotes = self._get_remote_parameters(ii)
            remotes.update({'start': start, 'end': end})

        # do results stuff here
        if self.return_all:
            msg = 'Returning all {0} results'.format(remotes['totalcount'])
        else:
            msg = 'Only returning the first {0} results.'.format(remotes['count'])

        if self.verbose:
            print('Results contain of a total of {0}. {1}'.format(remotes['totalcount'], msg))

        # get Marvin Results
        final = Results(**remotes)

        # get the final time
        posttime = datetime.datetime.now()
        self._final_time = (posttime - starttime)

        return final

    def _get_remote_parameters(self, interaction):
        ''' Retrieve or set parameters needed

        Parameters:
            interaction (object):
                The Marvin Interaction response object

        Returns:
            A dict of parameters needed to pass into Marvin Results

        '''

        results = interaction.results
        response_time = interaction.response_time
        assert isinstance(results, dict), 'Interaction results must be a dictionary'

        # get some parameters
        data = interaction.getData()
        params = results.get('params', None)
        query = results.get('query', self.search_filter)
        count = results.get('count', None)
        chunk = results.get('chunk', self.limit)
        totalcount = results.get('totalcount', None)
        runtime = results.get('runtime', None)

        # set some parameters when only data is available
        if len(results) == 1 and 'data' in results:
            # check first data row
            row = data[0]
            isstr = any([isinstance(i, six.string_types) for i in row])
            data = data[1:] if isstr else data
            # compute additional info
            params = row if isstr else []
            totalcount = len(data)
            count = totalcount
            chunk = self.limit
            runtime = response_time

        self.query = query
        self.params = params
        remotes = dict(response_time=response_time, params=params, query=query, results=data,
                       totalcount=totalcount, count=count, runtime=runtime, chunk=int(chunk),
                       mode=self.mode, queryobj=self)

        return remotes

    def _run_local(self, start=None, end=None, query_type=None):
        ''' Run a local database Query

        Parameters:
            start (int):
                A starting index when slicing the query
            end (int):
                An ending index when slicing the query
            query_type (str):
                The type of SQLAlchemy to submit. Can be "raw", "core", "orm"

        Returns:
            An instance of the :class:`~marvin.tools.query.results.Results`
            class containing the results of your Query.

        '''

        # Check for adding a sort
        self._sort_query()

        # Check to add the cache
        if self._caching:
            from marvin.core.caching_query import FromCache
            self.query = self.query.options(FromCache("default")).\
                options(*marvindb.cache_bits)

        # turn on streaming of results
        self.query = self.query.execution_options(stream_results=True)

        # set the start time of query
        starttime = datetime.datetime.now()

        # check for query and get count
        totalcount = self._get_query_count()

        # slice the query
        query = self._slice_query(start=start, end=end, totalcount=totalcount)

        # run the query and get the results
        results = self._get_results(query, query_type=query_type, totalcount=totalcount)

        # get the runtime
        endtime = datetime.datetime.now()
        self._run_time = (endtime - starttime)

        # convert to Marvin Results
        final = Results(results=results, query=query, count=self._count, mode=self.mode,
                        returntype=self.return_type, queryobj=self, totalcount=totalcount,
                        chunk=self.limit, runtime=self._run_time, start=self._start, end=self._end)

        # get the final time
        posttime = datetime.datetime.now()
        self._final_time = (posttime - starttime)

        return final

    def _sort_query(self):
        ''' Sort the SQLA query object by a given parameter '''

        if not isinstance(self.sort, type(None)):
            # check any shortcut names
            self.sort = self._marvinform._param_form_lookup.get_real_name(self.sort)
            
            # set the sort variable ModelClass parameter
            if '.' in self.sort:
                param = self.datamodel.parameters[str(self.sort)].full
            else:
                param = self.datamodel.parameters.get_full_from_remote(self.sort)
            sortparam = self._marvinform._param_form_lookup.mapToColumn(param)

            # check if sort param actually in the parameter list
            if sortparam.class_ not in self._modellist:
                return

            # If order is specified, then do the sort
            if self.order:
                assert self.order in ['asc', 'desc'], 'Sort order parameter must be either "asc" or "desc"'

                # Check if order by already applied
                if 'ORDER' in str(self.query.statement):
                    self.query = self.query.order_by(None)
                # Do the sorting
                if 'desc' in self.order:
                    self.query = self.query.order_by(desc(sortparam))
                else:
                    self.query = self.query.order_by(sortparam)

    def _get_query_count(self):
        ''' Get the SQL query count of rows

        First checks the query history table to look up if this query has
        already been run and a count produced.

        Returns:
            The total count of rows for the query

        '''

        totalcount = None
        if marvindb.isdbconnected:
            qm = self._check_history(check_only=True)
            totalcount = qm.count if qm else None

        # run count if it doesn't exist
        if totalcount is None:
            totalcount = self.query.count()

        return totalcount

    def _check_history(self, check_only=None, totalcount=None):
        ''' Check the query against the query history schema

        Looks up the current query in the query table of
        the history schema and if found, returns the SQLA object

        Parameters:
            check_only (bool):
                If True, only checks the history schema but does not write to it
            totalcount (int):
                The total count of rows to add, when adding a new query to the table

        Returns:
            The SQLAlchemy row from the query table of the history schema

        '''

        sqlcol = self._marvinform._param_form_lookup.mapToColumn('sql')
        stringfilter = self.search_filter.strip().replace(' ', '') if self.search_filter else ''
        rawsql = self.show().strip()
        returns = ','.join(self.return_params) if self.return_params else ''
        qm = self.session.query(sqlcol.class_).\
            filter(sqlcol == rawsql, sqlcol.class_.release == self.release).one_or_none()

        if check_only:
            return qm

        with self.session.begin():
            if not qm:
                qm = sqlcol.class_(searchfilter=stringfilter, n_run=1, release=self.release,
                                   count=totalcount, sql=rawsql, return_params=returns)
                self.session.add(qm)
            else:
                qm.n_run += 1

        return qm

    def _slice_query(self, start=None, end=None, totalcount=None):
        ''' Slice the SQLA query object

        Parameters:
            start (int):
                A starting index when slicing the query
            end (int):
                An ending index when slicing the query
            totalcount (int):
                The total count of rows of the query

        Returns:
            A new SQLA query object that has been sliced

        '''

        # get the new count if start and end exist
        if start and end:
            count = (end - start)
        else:
            count = totalcount

        # # run the query
        # res = self.query.slice(start, end).all()
        # count = len(res)
        # self.totalcount = count if not self.totalcount else self.totalcount

        # check history
        if marvindb.isdbconnected:
            __ = self._check_history(totalcount=totalcount)

        if count > self.count_threshold and self.return_all is False:
            # res = res[0:self.limit]
            start = 0
            end = self.limit
            count = (end - start)
            warnings.warn('Results contain more than {0} entries.  '
                          'Only returning first {1}'.format(self.count_threshold, self.limit), MarvinUserWarning)
        elif self.return_all is True:
            warnings.warn('Warning: Attempting to return all results. This may take a long time or crash.', MarvinUserWarning)
            start = None
            end = None
        elif start and end:
            warnings.warn('Getting subset of data {0} to {1}'.format(start, end), MarvinUserWarning)

        # slice the query
        query = self.query.slice(start, end)

        # set updated start, end, count, and total
        self._start = start
        self._end = end
        self._count = count
        self._total = totalcount

        return query

    def _get_results(self, query, query_type=None, totalcount=None):
        ''' Get the raw results of the query

        Runs the SQLAlchemy query.  query_type will determine how the query is run.
        "raw" means the query is run using the psycopg2 cursor object. "core" means
        the query is run using the SQLA connection object.  The "raw" and "core" methods
        will submit the raw sql string and retrieve the results in chunks using `fetchall`.
        orm" means the query is running using the SQLA query object.  This uses yield_per
        to generate the results in chunk.  It also folds similar strings together.

        Parameters:
            query (object):
                The current SQLA query object
            query_type (str):
                The type of SQLAlchemy to submit. Can be "raw", "core", "orm". Default is raw.
            totalcount (int):
                The total count of rows of the query

        Returns:
            A list of tupled results

        '''

        if query_type:
            assert query_type in ['raw', 'core', 'orm'], 'Query Type can only be raw, core, or orm.'
        else:
            query_type = 'raw'

        # run the query
        if query_type == 'raw':
            # use the db api cursor
            sql = str(self._get_sql(query))
            conn = marvindb.db.engine.raw_connection()
            cursor = conn.cursor('query_cursor')
            cursor.execute(sql)
            res = self._fetch_data(cursor)
            conn.close()
        elif query_type == 'core':
            # use the core connection
            sql = str(self._get_sql(query))
            with marvindb.db.engine.connect() as conn:
                results = conn.execution_options(stream_results=True).execute(sql)
                res = self._fetch_data(results)
        elif query_type == 'orm':
            # use the orm query
            yield_num = int(10**(np.floor(np.log10(totalcount))))
            results = string_folding_wrapper(query.yield_per(yield_num), keys=self.params)
            res = list(results)

        return res

    def _fetch_data(self, obj, n_rows=100000):
        ''' Fetch query results using fetchall or fetchmany

        Parameters:
            obj (object):
                SQLAlchemy connection object or Pyscopg2 cursor object
            n_rows (int):
                The number of rows to fetch at a time

        Returns:
            A list of results from a query

        '''

        res = []

        if not self.return_all:
            res = obj.fetchall()
        else:
            while True:
                rows = obj.fetchmany(n_rows)
                if rows:
                    res.extend(rows)
                else:
                    break
        return res

    @staticmethod
    def _get_sql(query):
        ''' Get the sql for a given query

        Parameters:
            query (object):
                An SQLAlchemy Query object

        Returns:
            A raw sql string
        '''

        return query.statement.compile(dialect=postgresql.dialect(), compile_kwargs={'literal_binds': True})

    def show(self, prop='query'):
        ''' Prints into the console

        Displays the query to the console with parameter variables plugged in.
        Works only in local mode.  Input prop can be one of query, joins, or filter.

        Allowed Values for Prop:
            - query: displays the entire query (default if nothing specified)
            - joins: displays the tables that have been joined in the query
            - filter: displays only the filter used on the query

        Parameters:
            prop (str):
                The type of info to print.  Can be 'query', 'joins', or 'filter'.

        Returns:
            The SQL string

        '''

        assert prop in [None, 'query', 'joins', 'filter'], 'Input must be query, joins, or filter'

        if self.data_origin == 'db':
            if not prop or prop == 'query':
                sql = self._get_sql(self.query)
            elif prop == 'joins':
                sql = self._joins
            elif prop == 'filter':
                if hasattr(self.query, 'whereclause'):
                    sql = self.query.whereclause.compile(dialect=postgresql.dialect(), compile_kwargs={'literal_binds': True})
                else:
                    sql = 'cannot extract filter from where clause'
            else:
                sql = self.__getattribute__(prop)

            return str(sql)
        elif self.data_origin == 'api':
            sql = self.search_filter
            return sql

    @classmethod
    def get_available_params(cls, paramdisplay='best', release=None):
        ''' Retrieve the available parameters to query on

        Retrieves a list of the available query parameters. Can either
        retrieve a list of all the parameters or only the vetted parameters.

        Parameters:
            paramdisplay (str {all|best}):
                String indicating to grab either all or just the vetted parameters.
                Default is to only return 'best', i.e. vetted parameters

        Returns:
            A list of all of the available queryable parameters
        '''
        assert paramdisplay in ['all', 'best'], 'paramdisplay can only be either "all" or "best"!'

        release = release or config.release
        if paramdisplay == 'all':
            if release not in datamodel:
                raise MarvinError('release {0} not found in query datamodel'.format(release))
            qparams = datamodel[release].groups.list_params('full')
        elif paramdisplay == 'best':
            qparams = query_params
        return qparams

    @remote_mode_only
    def save(self, path=None, overwrite=False):
        ''' Save the query as a pickle object

        Parameters:
            path (str):
                Filepath and name of the pickled object
            overwrite (bool):
                Set this to overwrite an existing pickled file

        Returns:
            path (str):
                The filepath and name of the pickled object

        '''

        sf = self.search_filter.replace(' ', '') if self.search_filter else 'anon'
        # set the path
        if not path:
            path = os.path.expanduser('~/marvin_query_{0}.mpf'.format(sf))

        # check for file extension
        if not os.path.splitext(path)[1]:
            path = os.path.join(path + '.mpf')

        path = os.path.realpath(path)

        if os.path.isdir(path):
            raise MarvinError('path must be a full route, including the filename.')

        if os.path.exists(path) and not overwrite:
            warnings.warn('file already exists. Not overwriting.', MarvinUserWarning)
            return

        dirname = os.path.dirname(path)
        if not os.path.exists(dirname):
            os.makedirs(dirname)

        # set bad pickled attributes to None
        attrs = ['session', 'datamodel', '_marvinform', '_modelgraph']

        # pickle the query
        try:
            with temp_setattr(self, attrs, None):
                pickle.dump(self, open(path, 'wb'), protocol=-1)
        except Exception as ee:
            if os.path.exists(path):
                os.remove(path)
            raise MarvinError('Error found while pickling: {0}'.format(str(ee)))

        return path

    @classmethod
    def restore(cls, path, delete=False):
        ''' Restore a pickled object

        Parameters:
            path (str):
                The filename and path to the pickled object

            delete (bool):
                Turn this on to delete the pickled fil upon restore

        Returns:
            Query (instance):
                The instantiated Marvin Query class
        '''
        obj = marvin_pickle.restore(path, delete=delete)
        obj._modelgraph = marvindb.modelgraph
        obj.session = marvindb.session
        obj.datamodel = datamodel[obj.release]
        obj.marvinform = obj.datamodel._marvinform
        return obj

    def update_return_params(self, params):
        ''' Update the list of return parameters '''

    #
    # This section describes the methods that run for local database queries
    #
    def _set_defaultparams(self):
        ''' Set the default parameters

        Loads any default parameters set.  Will also include appropriate
        defaults when specifying an object return type

        Default Parameters for Objects:
            - Cubes/RSS - plateifu/mangaid
            - Maps/ModelCube - plateifu/mangaid, bintype, template
            - Spaxel - plateifu/mangaid, x, y, bintype, template

        '''

        if self.return_type:
            assert self.return_type in ['cube', 'spaxel', 'maps', 'rss',
                                        'modelcube'], ('Query return_type must be either cube, spaxel, '
                                                       'maps, modelcube, rss')

        # set some initial defaults
        assert isinstance(self.default_params, (list, type(None))), 'default_params must be a list'
        defaults = self.default_params or (['cube.mangaid', 'cube.plateifu'] if self.nexus == 'cube' else [])

        extras = []
        if self.return_type == 'cube':
            extras = ['cube.mangaid', 'cube.plateifu']
        elif self.return_type == 'spaxel':
            pass
        elif self.return_type == 'modelcube':
            extras = ['bintype.name', 'template.name']
        elif self.return_type == 'rss':
            pass
        elif self.return_type == 'maps':
            extras = ['bintype.name', 'template.name']

        defaults.extend([e for e in extras if e not in defaults])
        self.default_params = defaults

        # add the defaults to the main set of parameters
        self.params.extend(self.default_params)

    def _set_return_params(self):
        ''' Set the return parameters '''

        # set the initial returns list
        returns = self.return_params or []
        returns = [returns] if not isinstance(returns, list) else returns

        # look up shortcut names for the return parameters
        full_returnparams = [self._marvinform._param_form_lookup._nameShortcuts[rp]
                             if rp in self._marvinform._param_form_lookup._nameShortcuts else rp
                             for rp in returns]

        self.return_params = full_returnparams

        # remove any return parameters that aren't already in the list of params
        use_only = self._filter_duplicates(full_returnparams)

        # add the return parameters to the main set of parameters
        self.params.extend(use_only)

    def _set_filter(self):
        ''' Set up the search filter '''

        # do nothing if nothing
        if not self.search_filter:
            return

        # check and parse the SQL string
        self._parse_sql_string()

    def _parse_sql_string(self):
        ''' Parse the SQL string '''

        # if params is a string, then parse and filter
        if not isinstance(self.search_filter, six.string_types):
            raise MarvinError('Input parameters must be a natural language string!')
        else:
            self._check_shortcuts_in_filter()
            try:
                parsed = parse_boolean_search(self.search_filter)
            except BooleanSearchException as e:
                raise MarvinError('Your boolean expression contained a syntax error: {0}'.format(e))

        # update the parameters dictionary
        self._check_parsed(parsed)
        self.filter_params.update(parsed.params)
        # remove keys that are already in the list of params
        filterkeys = self._filter_duplicates(parsed.uniqueparams)
        self.params.extend(filterkeys)

    def _check_shortcuts_in_filter(self):
        ''' Check for shortcuts in string filter and replace them '''

        # find all named parameters in the filter
        keys = re.findall('(?<!\d)[a-z\._]+\d*', self.search_filter)
        # remove the boolean operators
        keys = [i for i in keys if i not in ['and', 'or', 'not']]
        # look up real names of all keys
        real_names = {k: self._marvinform._param_form_lookup.get_real_name(k) for k in keys}

        # replace the shortcut names with real ones
        sf = self.search_filter
        for key, value in real_names.items():
            if key in sf:
                sf = sf.replace(key, value)
        self.search_filter = sf

    def _check_parsed(self, parsed):
        ''' Check the boolean parsed object

            check for function conditions vs normal.  This should be moved
            into SQLalchemy Boolean Search
        '''

        # Triggers for only one filter and it is a function condition
        if hasattr(parsed, 'fxn_name'):
            parsed.functions = [parsed]

        self._parsed = parsed

    def _filter_duplicates(self, columns, use_params='all'):
        ''' Filter out parameter duplicates

        Parameters:
            columns (list):
                A list of parameter names to check for duplicates in
            use_params (str):
                Indicates which parameter set to check the input columns against.
                Choices are "all" (default), "default", "return", "or filter".
                "all" uses self.params

        Returns:
            A list of new parameter names to add without existing duplicates
        '''

        assert use_params in ['all', 'default', 'return', 'filter'], (
            'Can only be "all", "default", "return", or "filter"')

        # get the params attribute
        if use_params == 'all':
            param_name = 'params'
        else:
            param_name = '{0}_params'.format(use_params)
        params = self.__getattribute__(param_name)

        # perform the check
        use_only = []
        for col in columns:
            # get the shortcut name
            shortcol = self._marvinform._param_form_lookup.get_shortcut_name(col)
            # if column or short name is already in list of params, add it
            if col not in params and shortcol not in params:
                # don't use the real column name for spaxelprop
                key = shortcol if 'spaxelprop' in col else col
                use_only.append(key)

        return use_only


    def _check_for(self, parameters, schema=None, tables=None, only=None):
        ''' Check if a schema or test of tables names are in the provided parameters

        Checks a list of parameters to see if any schema or table names are present

        Parameters:
            parameters (list)
                List of string parameters names to use in check
            schema (str):
                Schema name to check for in parameter list
            tables (list):
                List of table names to check for in parameter list
            only (bool):
                If True, checks if all parameters match the schema/table name conditions

        Returns:
            True if any/all of the parameters match the schema or tables

        '''

        # function to check if all or any parameters meet conditions
        fxn = all if only else any

        fparams = self._marvinform._param_form_lookup.mapToColumn(parameters)
        fparams = [fparams] if not isinstance(fparams, list) else fparams
        if schema and not tables:
            inschema = [schema in c.class_.__table__.schema for c in fparams]
            return True if fxn(inschema) else False
        if tables:
            schema_cond = schema if schema else ''
            tables = [tables] if not isinstance(tables, list) else tables
            # convert to full names
            tables = [self._marvinform._param_form_lookup._tableShortcuts.get(t, t) for t in tables]
            # get the parameter table names
            param_tables = [c.class_.__table__.name for c in fparams
                            if schema_cond in c.class_.__table__.schema]
            if only:
                diff = set(param_tables) ^ set(tables)
                return diff == set()
            else:
                intables = sum([[t in c for c in param_tables] for t in tables], [])
                return True if fxn(intables) else False

    def _build_query(self):
        ''' Build the query '''

        # create SQLA query parameters
        self._set_query_parameters()

        # create base SQLA object
        self._create_base_query()

        # join tables
        self._join_tables()

        # add condition
        self._add_condition()

        # add PipelineInfo
        self._add_pipeline()

        # check if the query filter is functional
        self._run_functional_queries()

        # check if the query parameters are against the DAP
        if self._check_for(self.params, schema='dapdb'):

            # Checks if the only table queried from dapdb is dapall. In that
            # case we allow the query.
            # checking for dapall, bintype, template since we have default parameters now
            all_dapall = self._check_for(self.params, schema='dapdb',
                                         tables=['dapall', 'bintype', 'template'], only=True)

            if not all_dapall and not config._allow_DAP_queries:
                raise NotImplementedError(
                    'DAP spaxel queries are disabled in this version. '
                    'We plan to reintroduce this feature in the future.')

            self._check_dapall_query()

    def _set_query_parameters(self):
        ''' Creates a list of database ModelClasses from a list of parameter names '''

        # adjust the default parameters for any necessary DAP
        self._add_default_params(['spaxelprop.x', 'spaxelprop.y', 'bintype.name', 'template.name'],
                                 tables=['spaxelprop', 'modelspaxel'])

        # adjust the default parameters for any necessary DAP
        self._add_default_params(['bintype.name', 'template.name'], tables=['dapall'])

        # adjust the default parameters for any necessary DAP
        self._add_default_params(['obsinfo.expnum', 'obsinfo.mgdpos'], tables=['obsinfo'])

        # final check to remove duplicates
        parset = set()
        psadd = parset.add
        self.params = [item for item in self.params if not (item in parset or psadd(item))]

        # create the list of parameter attributes
        queryparams = self._marvinform._param_form_lookup.mapToColumn(self.params)
        if not isinstance(queryparams, list):
            queryparams = [queryparams]

        # create a list of key names maintaining the column order
        self._query_params = [item for item in queryparams if item in set(queryparams)]
        self._query_params_order = [q.key for q in self._query_params]

    def _add_default_params(self, columns, tables=None):
        ''' Add new default parameters into the query

        Parameters:
            columns (list):
                A list of column names to add
            tables (list):
                A list of table names to check exists within the query
        '''

        # adjust the default parameters for any necessary columns
        if self._check_for(self.params, tables=tables):
            # remove any columns that are also defaults
            use_only = [c for c in columns if c not in self.default_params]
            self.default_params.extend(use_only)
            self.params.extend(use_only)

    def _create_base_query(self):
        ''' Create the base query session object.  Passes in a list of parameters defined in
            returnparams, filterparams, and defaultparams
        '''
        labeledqps = [qp.label(self.params[i]) for i, qp in enumerate(self._query_params)]
        self.query = self.session.query(*labeledqps)

    @update_config
    def _join_tables(self):
        ''' Build the join statement from the input parameters '''

        # from marvin import marvindb
        ifu = marvindb.datadb.IFUDesign

        self._joins = []

        # build list of SQLA models for the joins from the parameters
        self._modellist = []
        for param in self._query_params:
            # add the proper parameter Model
            if param.class_ not in self._modellist:
                self._modellist.append(param.class_)

            # if plateifu is a parameter, make sure we add the IFUDesign Model
            if 'plateifu' in str(param) and ifu not in self._modellist:
                self._modellist.append(ifu)

        # if there are no additional join tables, return
        if len(set(self._modellist)) == 1:
            return

        # Look up the nexus point.  If nexus is still none, takes the most common table from
        # the list of query parameters.  Default nexus is cube.
        nexus = self._marvinform.look_up_table(self.nexus)
        if not nexus:
            nexus = Counter(self._modellist).most_common(1)[0][0]

        # Gets the list of joins from ModelGraph. Uses Cube as nexus, so that
        # the order of the joins is the correct one.
        joinmodellist = self._modelgraph.getJoins(self._modellist, format_out='models', nexus=nexus)

        # Add the tables from the join list into the query
        for model in joinmodellist:
            name = '{0}.{1}'.format(model.__table__.schema, model.__tablename__)
            if not self._table_in_query(name):
                self._joins.append(model.__tablename__)
                if 'template' not in model.__tablename__:
                    self.query = self.query.join(model)
                else:
                    # assume template_kin only now, TODO deal with template_pop later
                    self.query = self.query.join(model, marvindb.dapdb.Structure.template_kin)

    def _table_in_query(self, name):
        ''' Checks if a given SQL table is already in the SQL query '''

        # do the check
        try:
            isin = name in str(self.query._from_obj[0])
        except IndexError as e:
            isin = False
        except AttributeError as e:
            if isinstance(self.query, six.string_types):
                isin = name in self.query
            else:
                isin = False
        return isin

    def _add_condition(self):
        ''' Loop over all input forms and add a filter condition based on the input parameter form data. '''

        # validate the forms
        self._validate_forms()

        # build the actual filter
        self._build_filter()

        # add the filter to the query
        if self.search_filter and not isinstance(self.filter, type(None)):
            self.query = self.query.filter(self.filter)

        # check for targets and quality flags to add in the filter
        self._check_targets()
        self._check_quality()

    def _validate_forms(self):
        ''' Validate all the data in the forms '''

        errors = []
        forms = self._set_forms()
        isgood = [form.validate() for form in forms.values()]
        if not all(isgood):
            inds = np.where(np.invert(isgood))[0]
            for index in inds:
                errors.append(list(forms.values())[index].errors)
            raise MarvinError('Parameters failed to validate: {0}'.format(errors))

    def _set_forms(self):
        ''' Set the appropriate WTForms in myforms and set the parameters '''

        forms = defaultdict(str)
        paramtree = tree()
        for key in self.filter_params.keys():
            forms[key] = self._marvinform.callInstance(self._marvinform._param_form_lookup[key], params=self.filter_params)
            paramtree[forms[key].Meta.model.__name__][key]
        return forms

    def _build_filter(self):
        ''' Builds a filter condition to load into sqlalchemy filter. '''

        # do nothing if nothing
        if not self.search_filter:
            return

        try:
            self.filter = self._parsed.filter(self._modellist)
        except BooleanSearchException as e:
            raise MarvinError('Your boolean expression could not me mapped to model: {0}'.format(e))

    def _check_targets(self):
        ''' Check for any target flags to add into the filter

        Checks for input manga_target flag labels in the Query.targets attribute
        and adds them to the Query filter condition.  All target labels are joined with
        a Boolean "or", then joined to the existing filter through a Boolean "and".

        To perform a bitwise not on a quality flag, specify the flag and value explicitly
        in the search_filter. E.g. "cube.manga_target1 & ~1024"

        Available target options are:
            primary - selects targets in both the PRIMARY_V1_2_0 and COLOR_ENHANCED_V1_2_0 samples
            color-enhanced - selects targets in the COLOR_ENHANCED_V1_2_0 sample
            secondary - selects targets in the SECONDARY_V1_2_0 sample
            ancillary - selects any targets in the ANCILLARY sample
            stellar libraray - selects targets from the stellar library
            flux standards - selects flux standard stars
            any ancillary catalog name, e.g. 'MWA', or 'DWARF'

        Example
            >>> # perform query selecting primary and secondary targets
            >>> q = Query(targets=['primary', 'secondary'])
            >>>
            >>> # select ancillary targets from Milky Way Analogs and Dwarfs
            >>> q = Query(targets=['MWA', 'DWARF'])
            >>>
            >>> # select all targets that are not a part of the PRIMARY_v1_2_0 sample
            >>> q = Query(search_filter='cube.manga_target & ~1024')
            >>>
        '''

        if not self.targets:
            return

        targets = [self.targets] if not isinstance(self.targets, list) else self.targets
        targets = [t.upper() for t in targets]

        # check for ancillary targets
        ancillaries = [a.upper() for a in self.datamodel.bitmasks['MANGA_TARGET3'].schema.label]
        anc_labels = list(set(targets) & set(ancillaries))
        targets = list(set(targets) - set(ancillaries))

        # build the string filter
        target_filter = ''
        for target in targets:
            target_filter = self._create_target_filter(target, target_filter=target_filter)

        # add in any ancillary targets
        if anc_labels:
            target_filter = self._create_target_filter('ancillary', target_filter=target_filter, ancillaries=anc_labels)

        # parse the filter and add to the main
        self._add_filter(target_filter)

    def _create_target_filter(self, target, target_filter='', ancillaries=None):
        ''' Create a string target bitwise filter

        Creates a string manga_target filter bitwise filter.  If passed an existing
        target_filter string, it will append the new filter as an "or" boolean condition.
        Example filter syntax is 'cube.manga_target1 & 5120'

        Parameters:
            target (str):
                The name of the target sample to load
            target_filter (str):
                The current existing target filter string
            ancillarites (list):
                A list of ancillary target labels

        Returns:
            A string filter condition

        '''
        target = target.lower()
        defaults = ['primary', 'color-enhanced', 'secondary', 'ancillary', 'stellar library', 'flux standards']
        #options = defaults + ancillaries
        assert target in defaults, 'target list can only contain one of {0}'.format(defaults)

        if 'primary' == target:
            name = 'manga_target1'
            value = self.datamodel.bitmasks['MANGA_TARGET1'].labels_to_value(['PRIMARY_v1_2_0', 'COLOR_ENHANCED_v1_2_0'])
        elif 'color-enhanced' == target:
            name = 'manga_target1'
            value = self.datamodel.bitmasks['MANGA_TARGET1'].labels_to_value(['COLOR_ENHANCED_v1_2_0'])
        elif 'secondary' == target:
            name = 'manga_target1'
            value = self.datamodel.bitmasks['MANGA_TARGET1'].labels_to_value(['SECONDARY_v1_2_0'])
        elif 'ancillary' == target:
            name = 'manga_target3'
            value = self.datamodel.bitmasks['MANGA_TARGET3'].labels_to_value(ancillaries) if ancillaries else 0
        elif 'stellar library' == target:
            name = 'manga_target2'
            value = sum([1<<i for i in (self.datamodel.bitmasks['MANGA_TARGET2'].schema.bit[2:17])])
        elif 'flux standards' == target:
            name = 'manga_target2'
            bits = [20, 22, 23, 25, 26, 27]
            value = sum([1<<i for i in bits])

        # add the column to the query
        self._add_columns('cube.{0}'.format(name))

        base = ' or ' if target_filter else ''
        op = '>' if value == 0 else '&'
        target_filter += '{0}cube.{1} {2} {3}'.format(base, name, op, value)
        return target_filter

    def _check_quality(self):
        ''' Check for any quality flags to add into the filter

        Checks for input DRP or DAP quality flag labels in the Query.quality attribute
        and adds them to the Query filter condition.  All quality labels are joined with
        a Boolean "or", then joined to the existing filter through a Boolean "and".

        To perform a bitwise not on a quality flag, specify the flag and value explicitly
        in the search_filter. E.g. "cube.quality & ~64"

        Example
            >>> # perform query selecting galaxies with a bad flux calibration
            >>> q = Query(quality=['BADFLUX'])
            >>>
            >>> # select galaxies that do not have bad flux calibration
            >>> q = Query(search_filter='cube.quality & ~256')
            >>>

        '''

        if not self.quality:
            return

        # format the quality flags
        quality = [self.quality] if not isinstance(self.quality, list) else self.quality
        quality = [t.upper() for t in quality]

        # get all the available flags
        flags = sum([list(i.schema.label) for i in self.datamodel.bitmasks.values()
                     if 'QUAL' in i.name or 'MASK' in i.name], [])
        assert set(quality).issubset(set(flags)), 'quality flag must be one of {0}'.format(flags)

        # get any individual label sets
        drplabels = self._get_labelset(quality, name='MANGA_DRP3QUAL')
        daplabels = self._get_labelset(quality, name='MANGA_DAPQUAL')
        dapspeclabels = self._get_labelset(quality, name='MANGA_DAPSPECMASK')
        dappixlabels = self._get_labelset(quality, name='MANGA_DAPPIXMASK')

        # build the quality filter
        quality_filter = ''
        if drplabels:
            quality_filter = self._create_quality_filter(drplabels, flag='DRP3QUAL', quality_filter=quality_filter)
        if daplabels:
            quality_filter = self._create_quality_filter(daplabels, flag='DAPQUAL', quality_filter=quality_filter)

        # parse the filter and add to the main
        if quality_filter:
            spaxelprop = marvindb.dapdb.__getattribute__('Clean{0}'.format(self.datamodel.dap_datamodel.property_table))
            models = [marvindb.datadb.Cube, marvindb.dapdb.File, spaxelprop]
            self._add_filter(quality_filter, modellist=models)

    def _get_labelset(self, flags, name=None):
        ''' Return matching labels in the set of flags

        Selects out those labels that are in the set of labels from the
        named flag.

        Parameters:
            flags (list):
                A list of the input flag labels
            name (str):
                The name of the flag set to match against

        Returns:
            The list of labels that are in the named flag set
        '''

        # return empty list if the bitmask set not available
        if name not in self.datamodel.bitmasks:
            return []

        labels = [q.upper() for q in self.datamodel.bitmasks[name].schema.label]
        flags = [f.upper() for f in flags]
        inset = list(set(flags) & set(labels))
        return inset

    def _create_quality_filter(self, labels, flag='DRP3QUAL', quality_filter=''):
        ''' Create a string quality bitwise filter

        Creates a string quality filter bitwise filter.  If passed an existing
        quality_filter string, it will append the new filter as an "or" boolean condition.
        Example filter syntax is 'cube.quality & 64'

        Parameters:
            labels (list):
                A list of string quality flag labels
            flag (str):
                The short hand name of the flag, i.e. MANGA_XXXX
            quality_filter(str):
                An existing string quality filter condition

        Returns:
            A string quality filter condition

        '''

        # get the value given the labels
        name = 'MANGA_{0}'.format(flag.upper())
        value = self.datamodel.bitmasks[name].labels_to_value(labels)

        # add the column to the query
        colname = 'cube.quality' if 'DRP' in flag else 'file.quality'
        self._add_columns(colname)

        base = ' or ' if quality_filter else ''
        quality_filter += '{0}{1} & {2}'.format(base, colname, value)
        return quality_filter

    def _add_filter(self, strfilter, modellist=None):
        ''' Parse and add a string filter into the query

        Pass a string filter condition into the boolean parser and
        explicitly add it to the query filter

        Parameters:
            strfilter (str):
                The filter condition to parse
            modellist (list):
                A list of models needed for the filter to
                identify the correct parameters

        '''

        modellist = modellist if modellist else marvindb.datadb
        # parse the filter and add to the main
        parsed = parse_boolean_search(strfilter)
        f = parsed.filter(modellist)
        self.query = self.query.filter(and_(f))

    def _add_columns(self, columns):
        ''' Add columns to the query

        Parameters:
            columns (list):
                A list of string column names to add to the query
        '''

        # get new columns not already added
        columns = [columns] if not isinstance(columns, list) else columns
        new_columns = list(set(columns) - set(self.params))
        if any(new_columns):
            colattrs = self._marvinform._param_form_lookup.mapToColumn([c for c in new_columns])
            colattrs = [colattrs] if not isinstance(colattrs, list) else colattrs
            self.query = self.query.add_columns(*colattrs)
            self.params.extend(new_columns)

    def _add_pipeline(self):
        ''' Adds the DRP and DAP Pipeline Info into the Query '''

        self._drp_alias = aliased(marvindb.datadb.PipelineInfo, name='drpalias')
        self._dap_alias = aliased(marvindb.datadb.PipelineInfo, name='dapalias')

        drppipe = self._get_pipe_info('drp')
        dappipe = self._get_pipe_info('dap')

        # Add DRP pipeline version
        if drppipe:
            self.query = self.query.join(self._drp_alias, marvindb.datadb.Cube.pipelineInfo).\
                filter(self._drp_alias.pk == drppipe.pk)

        # Add DAP pipeline version
        if dappipe:
            self.query = self.query.join(self._dap_alias, marvindb.dapdb.File.pipelineinfo).\
                filter(self._dap_alias.pk == dappipe.pk)

    def _get_pipe_info(self, pipename):
        ''' Retrieve the pipeline Info for a given pipeline version name '''

        assert pipename.lower() in ['drp', 'dap'], 'Pipeline Name must either be DRP or DAP'

        # bindparam values
        bindname = 'drpver' if pipename.lower() == 'drp' else 'dapver'
        bindvalue = self._drpver if pipename.lower() == 'drp' else self._dapver

        # class names
        if pipename.lower() == 'drp':
            inclasses = self._table_in_query('cube') or 'cube' in str(self.query.statement.compile())
        elif pipename.lower() == 'dap':
            inclasses = self._table_in_query('file') or 'file' in str(self.query.statement.compile())

        # set alias
        pipealias = self._drp_alias if pipename.lower() == 'drp' else self._dap_alias

        # get the pipeinfo
        if inclasses:
            pipeinfo = marvindb.session.query(pipealias).\
                join(marvindb.datadb.PipelineName, marvindb.datadb.PipelineVersion).\
                filter(marvindb.datadb.PipelineName.label == pipename.upper(),
                       marvindb.datadb.PipelineVersion.version == bindparam(bindname, bindvalue)).one()
        else:
            pipeinfo = None

        return pipeinfo

    def _group_by(self, params=None):
        ''' Group the query by a set of parameters

        Parameters:
            params (list):
                A list of string parameter names to group the query by

        Returns:
            A new SQLA Query object
        '''

        if not params:
            params = [d for d in self.default_params if 'spaxelprop' not in d]

        #newdefaults = self._marvinform._param_form_lookup.mapToColumn(params)
        newdefaults = [d for d in self._query_params if str(d).lower() in params]
        self.default_params = params
        self.params = params
        newq = self.query.from_self(*newdefaults).group_by(*newdefaults)
        return newq

    def _check_query(self, name):
        ''' Check if string is inside the query statement '''

        qstate = str(self.query.statement.compile(compile_kwargs={'literal_binds': True}))
        return name in qstate

    def _update_filter_params(self, param):
        ''' Update the input parameters '''
        param = {key: val.decode('UTF-8') if '*' not in val.decode('UTF-8') else
                 val.replace('*', '%').decode('UTF-8') for key, val in param.items()
                 if key in self.filter_params.keys()}
        self.filter_params.update(param)

    def _already_in_filter(self, names):
        ''' Checks if the parameter name already added into the filter '''

        infilter = None
        if names:
            if not isinstance(self.query, type(None)):
                if not isinstance(self.query.whereclause, type(None)):
                    wc = str(self.query.whereclause.compile(dialect=postgresql.dialect(),
                             compile_kwargs={'literal_binds': True}))
                    infilter = any([name in wc for name in names])

        return infilter

    #
    # Methods specific to functional queries
    #
    def _run_functional_queries(self):
        ''' Checks for functional filter conditions and runs them '''

        if not self.search_filter:
            return

        # check for additional modifier criteria
        if self._parsed.functions:
            # loop over all functions
            for fxn in self._parsed.functions:
                # look up the function name in the marvinform dictionary
                try:
                    methodname = self._marvinform._param_fxn_lookup[fxn.fxn_name]
                except KeyError as e:
                    raise MarvinError('Could not set function: {0}'.format(e))
                else:
                    # run the method
                    methodcall = self.__getattribute__(methodname)
                    methodcall(fxn)

    def _get_good_spaxels(self):
        ''' Subquery - Counts the number of good spaxels

        Counts the number of good spaxels with binid != -1
        Uses the spaxelprop.bindid_pk != 9999 since this is known and set.
        Removes need to join to the binid table

        Returns:
            bincount (subquery):
                An SQLalchemy subquery to be joined into the main query object
        '''

        spaxelname = self._spaxelclass.__name__
        bincount = self.session.query(self._spaxelclass.file_pk.label('binfile'),
                                      func.count(self._spaxelclass.pk).label('goodcount'))

        # optionally add the filter if the table is SpaxelProp
        if 'CleanSpaxelProp' not in spaxelname:
            bincount = bincount.filter(self._spaxelclass.binid != -1)

        # group the results by file_pk
        bincount = bincount.group_by(self._spaxelclass.file_pk).subquery('bingood', with_labels=True)

        return bincount

    def _get_count_of(self, expression):
        ''' Subquery - Counts spaxels satisfying an expression

        Counts the number of spaxels of a given
        parameter above a certain value.

        Parameters:
            expression (str):
                The filter expression to parse

        Returns:
            valcount (subquery):
                An SQLalchemy subquery to be joined into the main query object

        Example:
            >>> expression = 'spaxelprop.emline_gflux_ha_6564 >= 25'
        '''

        # parse the expression into name, operator, value
        param, ops, value = self._parse_expression(expression)
        # look up the InstrumentedAttribute, Operator, and convert Value
        attribute = self._marvinform._param_form_lookup.mapToColumn(param)
        op = opdict[ops]
        value = float(value)
        # Build the subquery
        valcount = self.session.query(self._spaxelclass.file_pk.label('valfile'),
                                      (func.count(self._spaxelclass.pk)).label('valcount')).\
            filter(op(attribute, value)).\
            group_by(self._spaxelclass.file_pk).subquery('goodhacount', with_labels=True)

        return valcount

    def _get_percent(self, fxn, **kwargs):
        ''' Query - Computes count comparisons

        Retrieves the number of objects that have satisfy a given expression
        in x% of good spaxels.  Expression is of the form
        Parameter Operand Value. This function is mapped to
        the "npergood" filter name.

        Syntax: fxn_name(expression) operator value

        Parameters:
            fxn (str):
                The function condition used in the query filter

        Example:
            >>> fxn = 'npergood(spaxelprop.emline_gflux_ha_6564 > 25) >= 20'
            >>> Syntax: npergood() - function name
            >>>         npergood(expression) operator value
            >>>
            >>> Select objects that have Ha flux > 25 in more than
            >>> 20% of their (good) spaxels.
        '''

        # get the appropriate SpaxelProp ModelClass
        self._spaxelclass = self._marvinform._param_form_lookup['spaxelprop.file'].Meta.model

        # parse the function into name, condition, operator, and value
        name, condition, ops, value = self._parse_fxn(fxn)
        percent = float(value) / 100.
        op = opdict[ops]

        # Retrieve the necessary subqueries
        bincount = self._get_good_spaxels()
        valcount = self._get_count_of(condition)

        # Join to the main query
        self.query = self.query.join(bincount, bincount.c.binfile == self._spaxelclass.file_pk).\
            join(valcount, valcount.c.valfile == self._spaxelclass.file_pk).\
            filter(op(valcount.c.valcount, percent * bincount.c.goodcount))

        # Group the results by main default datadb parameters, so as not to include all spaxels
        newdefs = [d for d in self.default_params if 'spaxelprop' not in d]
        self.query = self._group_by(params=newdefs)

    def _parse_fxn(self, fxn):
        ''' Parse a fxn condition '''
        return fxn.fxn_name, fxn.condition, fxn.operator, fxn.value

    def _parse_expression(self, expr):
        ''' Parse an expression '''
        return expr.fullname, expr.op, expr.value

    def _check_dapall_query(self):
        ''' Checks if the query is on the DAPall table, and regroup the parameters plateifu'''

        isdapall = self._check_query('dapall')
        # if isdapall:
        #     self.query = self._group_by()

    def _radial_query(self, fxn, **kwargs):
        ''' Runs a radial cone search around an RA, Dec

        Performs a radial cone search around an RA, Dec point
        within some specified radial distance in units of degrees.

        Syntax: radial(ra, dec, radius)

        Parameters:
            fxn (str):
                The function condition used in the query filter

        Example:
            >>> fxn = 'radial(232.5447, 48.6902, 1)'
            >>> Syntax: radial() - function name
            >>>         radial(ra, dec, radius)
            >>>
            >>> Select objects that are within 1 degree of
            >>> RA, Dec = (232.5447, 48.6902)

        '''

        # extract the RA, Dec and search radius
        ra, dec = map(float, (fxn.coords))
        radius = float(fxn.value)

        # add RA, Dec as returned columns
        self._add_columns(['cube.ra', 'cube.dec'])

        # Join to the main query
        cone_filter = func.q3c_radial_query(marvindb.datadb.Cube.ra, marvindb.datadb.Cube.dec, ra, dec, radius)
        self.query = self.query.filter(cone_filter)



# !usr/bin/env python
# -*- coding: utf-8 -*-
#
# Modified from mangahi.py by Karen Masters.

from __future__ import print_function, division, absolute_import

import astropy
import marvin.tools

from .base import VACMixIn


class GZVAC(VACMixIn):
    """Provides access to the MaNGA Galaxy Zoo Morphology VAC.

    VAC name: MaNGA Morphologies from Galaxy Zoo

    URL: https://www.sdss.org/dr17/data_access/value-added-catalogs/?vac_id=manga-morphologies-from-galaxy-zoo
    
    Description Returns Galaxy Zoo morphology for MaNGA galaxies. 
    The Galaxy Zoo (GZ) data for SDSS galaxies has been split over several iterations of www.galaxyzoo.org, 
    with the MaNGA target galaxies being spread over five different GZ data sets. In this value added catalog,
    for DR15, we bring all of these galaxies into one single catalog and re-run the debiasing code (Hart et al. 2016) in 
    a consistent manner across the all the galaxies. This catalog includes data from Galaxy Zoo 2 (previously 
    published in Willett et al. 2013) and newer data from Galaxy Zoo 4 (currently unpublished).

    For DR17, we provide new and updated Galaxy Zoo (GZ) data for the final MaNGA galaxies. This has been split over 
    three files, each corresponding to a separate GZ catalogue. We have `MaNGA_GZD_auto-v1_0_1.fits`, which
    corresponds to the automated classifications GZ DECaLS, described in Walmsley et al. 2021. There is also 
    `MaNGA_gzUKIDSS-v1_0_1.fits`, which correponds to GZ:UKIDSS. Finally, we have put the rest of GZ 
    (so not including GZ DECaLS and GZ:UKIDSS) in `MaNGA_gz-v2_0_1.fits`. For more information, please 
    refer to the datamodels provided.

    Authors: Coleman Krawczyk, Karen Masters, Tobias Géron and the rest of the Galaxy Zoo Team.

    """

    # Required parameters
    name = "galaxyzoo"
    description = "Returns Galaxy Zoo morphology"
    version = {"MPL-7": "v1_0_1", "MPL-8": "v1_0_1", "DR15": "v1_0_1", "DR16": "v1_0_1", 
               "MPL-11" : None, "DR17" : None}
    display_name = 'Galaxy Zoo'
    url = 'https://www.sdss.org/dr17/data_access/value-added-catalogs/?vac_id=manga-morphologies-from-galaxy-zoo'

    # optional Marvin Tools to attach your vac to
    include = (marvin.tools.cube.Cube, marvin.tools.maps.Maps, marvin.tools.modelcube.ModelCube)

    # Required method
    def set_summary_file(self, release):
        ''' Sets the path to the GalaxyZoo summary file.
        
        Sets the paths to the GalaxyZoom summary file(s).  For DR15 this is a single summary file,
        while for DR17, this has been split into three files, so ``self.summary_file`` and 
        ``self.path_params`` return lists for DR17.
        '''

        self.summary_file = []
        self.path_params = []

        # define the variables to build a unique path to your VAC file
        if release in ["DR17", "MPL-11"]:  
            # for DR17, path is more complicated as we have three files.
            files = ['GZD_auto', 'gzUKIDSS', 'gz']
            version_DR17 = {"GZD_auto": "v1_0_1", "gzUKIDSS": "v1_0_1", "gz": "v2_0_1"}
            for file in files:
                params = {"file" : file, "ver" : version_DR17[file]}
                self.path_params.append(params)
                
                # get_path returns False if the files do not exist locally
                self.summary_file.append(self.get_path("mangagalaxyzoo", path_params=params))
            self.data_container = dict(zip(files, self.summary_file))
        else: 
            # for other releases prior to DR17, simply do based on release as before
            self.path_params.append({"ver": self.version[release], "file": "gz"})
            self.summary_file.append(self.get_path("mangagalaxyzoo", path_params=self.path_params[0]))

    # Required method
    def get_target(self, parent_object):
        ''' Accesses VAC data for a specific target from a Marvin Tool object '''

        # get any parameters you need from the parent object
        mangaid = parent_object.mangaid

        # download the vac from the SAS if it does not already exist locally
        for i in range(len(self.summary_file)):
            if not self.file_exists(self.summary_file[i]):
                self.summary_file[i] = self.download_vac("mangagalaxyzoo", path_params=self.path_params[i])
                
        # Open the file(s) using fits.getdata for extension 1
        data = []
        for i in range(len(self.summary_file)):
            data.append(astropy.io.fits.getdata(self.summary_file[i], 1))

        # Return selected line(s)
        result = {}
        keys = ['gz_auto','gz_ukidss', 'gz']

        for i in range(len(self.summary_file)):
            indata = mangaid in data[i]["mangaid"]
            if not indata:
                pass # passes instead of adding an empty key-value pair.

            else:
                idx = data[i]["mangaid"] == mangaid
                result[keys[i]] = data[i][idx]

        # return the result
        if len(self.summary_file) == 1:
            # to make sure output stays the same if version is < DR17
            return list(result.values())[0]
        else:
            # for DR17, return dict with classifications of all three files
            return result 





#!/usr/bin/env python
# encoding: utf-8
#
# maskbit.py
#
# @Author: Brett Andrews <andrews>
# @Date:   2017-10-06 17:10:00
# @Last modified by:   andrews
# @Last modified time: 2017-11-15 10:11:29

from __future__ import division, print_function, absolute_import

import pandas as pd

from marvin.utils.general.maskbit import Maskbit

__ALL__ = ('get_maskbits')


def get_maskbits(release):
    maskbits = {
        'MANGA_TARGET1': Maskbit('MANGA_TARGET1', description='Targeting bits for all galaxy targets.'),
        'MANGA_TARGET2': Maskbit('MANGA_TARGET2', description='Targeting bits for all non-galaxy targets.'),
        'MANGA_TARGET3': Maskbit('MANGA_TARGET3', description='Targeting bits for ancillary targets.'),
        'MANGA_DRP3QUAL': Maskbit('MANGA_DRP3QUAL', description='Describes the quality of the final cubes and RSS files.'),
        'MANGA_DAPQUAL': Maskbit('MANGA_DAPQUAL', description='Describes the overall quality of the data.'),
        'MANGA_DRP3PIXMASK': Maskbit('MANGA_DRP3PIXMASK', description='Describes whether a given spaxel should be used for science analyses.'),
        'MANGA_DAPPIXMASK': Maskbit('MANGA_DAPPIXMASK',
                                    description='2d image bitmap used to describe the quality of '
                                    'individual pixel measurements in the DAP MAPS file.'),
        'MANGA_DAPSPECMASK': Maskbit('MANGA_DAPSPECMASK',
                                     description='3d cube bitmap used to describe the quality of '
                                     'individual spaxel fits in the DAP model data cube file.')
        }

    if release == 'MPL-4':
        maskbits['MANGA_DAPPIXMASK'] = MPL4_dappixmask()
        __ = maskbits.pop('MANGA_DAPQUAL')

    return maskbits


def MPL4_dappixmask():
    schema = pd.DataFrame([(0, 'DONOTUSE', 'Do not use this spaxel for science')],
                          columns=['bit', 'label', 'description'])
    return Maskbit(name='MANGA_DAPPIXMASK', schema=schema,
                   description='2d image bitmap used to describe the quality of individual pixel '
                               'measurements in the DAP MAPS file.')




import contextlib
import functools
import os
import re
from copy import deepcopy

from brain.utils.general import build_routemap
from flask import current_app
from flask_jwt_extended import fresh_jwt_required
from webargs import ValidationError, fields, validate
from webargs.flaskparser import parser, use_args, use_kwargs

from marvin import config
from marvin.core.exceptions import MarvinError
from marvin.utils.datamodel.dap import datamodel as dm
from marvin.utils.general import validate_jwt


def plate_in_range(val):
    if int(val) < 6500:
        raise ValidationError('Plateid must be > 6500')


# List of global View arguments across all API routes
viewargs = {'name': fields.String(required=True, location='view_args', validate=[validate.Length(min=4),
                                  validate.Regexp('^[0-9-]*$')]),
            'galid': fields.String(required=True, location='view_args', validate=[validate.Length(min=4),
                                   validate.Regexp('^[0-9-]*$')]),
            'bintype': fields.String(required=True, location='view_args'),
            'template': fields.String(required=True, location='view_args'),
            'property_name': fields.String(required=True, location='view_args'),
            'channel': fields.String(required=True, location='view_args'),
            'binid': fields.Integer(required=True, location='view_args', validate=validate.Range(min=-1, max=5800)),
            'plateid': fields.String(required=True, location='view_args', validate=[validate.Length(min=4, max=5),
                                     plate_in_range]),
            'x': fields.Integer(required=True, location='view_args', validate=validate.Range(min=0, max=100)),
            'y': fields.Integer(required=True, location='view_args', validate=validate.Range(min=0, max=100)),
            'mangaid': fields.String(required=True, location='view_args', validate=validate.Length(min=4, max=20)),
            'paramdisplay': fields.String(required=True, location='view_args', validate=validate.OneOf(['all', 'best'])),
            'cube_extension': fields.String(required=True, location='view_args',
                                            validate=validate.OneOf(['flux', 'ivar', 'mask',
                                                                     'disp', 'predisp',
                                                                     'specres', 'specresd',
                                                                     'prespecres',
                                                                     'prespecresd'])),
            'modelcube_extension': fields.String(required=True, location='view_args',
                                                 validate=validate.OneOf(['flux', 'ivar', 'mask',
                                                                          'model', 'emline',
                                                                          'emline_base',
                                                                          'emline_mask'])),
            'colname': fields.String(required=True, location='view_args', allow_none=True),
            'fiberid': fields.Integer(required=True, location='view_args',
                                      validate=validate.Range(min=-1, max=5800)),
            }

# List of all form parameters that are needed in all the API routes
# allow_none = True allows for the parameter to be non-existent when required=False and missing is not set
# setting missing = None by itself also works except when the parameter is also required
# (i.e. required=True + missing=None does not trigger the "required" validation error when it should)
params = {'query': {'searchfilter': fields.String(allow_none=True),
                    'paramdisplay': fields.String(allow_none=True, validate=validate.OneOf(['all', 'best'])),
                    'task': fields.String(allow_none=True, validate=validate.OneOf(['clean', 'getprocs'])),
                    'start': fields.Integer(allow_none=True, validate=validate.Range(min=0)),
                    'end': fields.Integer(allow_none=True, validate=validate.Range(min=0)),
                    'offset': fields.Integer(allow_none=True, validate=validate.Range(min=0)),
                    'limit': fields.Integer(allow_none=True, missing=100, validate=validate.Range(max=50000)),
                    'sort': fields.String(allow_none=True),
                    'order': fields.String(missing='asc', validate=validate.OneOf(['asc', 'desc'])),
                    'rettype': fields.String(allow_none=True, validate=validate.OneOf(['cube', 'spaxel', 'maps', 'rss', 'modelcube'])),
                    'returnparams': fields.DelimitedList(fields.String(), allow_none=True),
                    'defaults': fields.DelimitedList(fields.String(), allow_none=True),
                    'targets': fields.DelimitedList(fields.String(), allow_none=True),
                    'quality': fields.DelimitedList(fields.String(), allow_none=True),
                    'return_all': fields.Boolean(allow_none=True),
                    'format_type': fields.String(allow_none=True, validate=validate.OneOf(['list', 'listdict', 'dictlist'])),
                    'caching': fields.Boolean(allow_none=True),
                    'query_type': fields.String(allow_none=True, validate=validate.OneOf(['raw', 'core', 'orm']))
                    },
          'search': {'searchbox': fields.String(required=True),
                     'parambox': fields.DelimitedList(fields.String(), allow_none=True)
                     },
          'index': {'galid': fields.String(allow_none=True, validate=[validate.Length(min=4), validate.Regexp('^[0-9-]*$')]),
                    'mplselect': fields.String(allow_none=True, validate=validate.Regexp('^(MPL-|DR)([0-9]{1,2}$)'))
                    },
          'galaxy': {'plateifu': fields.String(allow_none=True, validate=validate.Length(min=8, max=11)),
                     'toggleon': fields.String(allow_none=True, validate=validate.OneOf(['true', 'false'])),
                     'image': fields.Url(allow_none=True),
                     'imheight': fields.Integer(allow_none=True, validate=validate.Range(min=0, max=1000)),
                     'imwidth': fields.Integer(allow_none=True, validate=validate.Range(min=0, max=1000)),
                     'type': fields.String(allow_none=True, validate=validate.OneOf(['optical', 'heatmap'])),
                     'x': fields.String(allow_none=True),
                     'y': fields.String(allow_none=True),
                     'mousecoords[]': fields.List(fields.String(), allow_none=True),
                     'bintemp': fields.String(allow_none=True),
                     'params[]': fields.List(fields.String(), allow_none=True)
                     },
          'explore': {'target': fields.String(required=True),
                      'mapchoice': fields.String(allow_none=True),
                      'btchoice': fields.String(allow_none=True)}
          }


# Add a custom Flask session location handler
@parser.location_handler('session')
def parse_session(req, name, field):
    from flask import session as current_session
    value = current_session.get(name, None)
    return value


class ArgValidator(object):
    ''' Web/API Argument validator

    This class controls the parameter / argument validation for all API and Web route
    parameters, both in the views (url names), and in the request body (get/post parameters).
    This uses the Python package webargs and marshmallow to define and control all
    validation of parameters. See `webargs <https://webargs.readthedocs.io/en/latest/>`_ and
    `marshmallow <https://marshmallow.readthedocs.io/en/latest/>`_.

    There are two global dictionaries defining the argument validations.  `view_args` defines the
    list of all named arguments in the api/web Flask routes, e.g. `/marvin/api/cubes/<name>/` or
    `/marvin/api/maps/<name>/<bintype>/<template>/`.   `params` defines the list of all parameters
    used in routes as form arguments in GET or POST requests, in both the web and api Flask routes.
    Each entry is a dictionary of {page: dict of parameter validators}.  For example, the `query` key contains
    parameters related to sending requests to the API or Web Query views.

    You can use the validator in two ways, as a decorator, or manually inside a function.
    To use the validator as a decorator, decorate your route with the `check_args` method.  Your
    route function must contain an args input variable which is a dictionary of the validated
    parameters as checked by the decorator. See :meth:`marvin.api.ArgValidator.check_args` for more info.

    To use the validator manually, call the `manual_parse` method inside the function. See
    :meth:`marvin.api.ArgValidator.manual_parse` for more info.

    Example:
        >>> # use as a decorator (checks any view url arguments)
        >>> from marvin.api.base import arg_validate as av
        >>> @route('/myroute/')
        >>> @av.check_args()
        >>> def do_stuff(args):
        >>>     return stuff
        >>>
        >>> # use manually inside a Flask view function (checks the galaxy request parameters
        >>> # and required plateifu is set)
        >>> from flask import request
        >>> def do_stuff(self):
        >>>     args = av.manual_parse(self, request, use_params='galaxy', required='plateifu')
        >>>     return stuff

    '''

    def __init__(self, urlmap={}):
        self.release = None
        self.endpoint = None
        self.dapver = None
        self.urlmap = urlmap
        self.base_args = {'release': fields.String(required=True,
                          validate=validate.Regexp('^(MPL-|DR)([0-9]{1,2}$)'))}
        self.use_params = None
        self._required = None
        self._setmissing = None
        self._main_kwargs = {}
        self.final_args = {}
        self.final_args.update(self.base_args)

        self._parser = parser
        self.use_args = use_args
        self.use_kwargs = use_kwargs

    def _check_urlmap(self):
        ''' Check that urlmap exists and build it if not '''
        if not self.urlmap:
            urlmap = build_routemap(current_app)
            self.urlmap = urlmap

    def _reset_final_args(self):
        ''' Resets the final args dict '''
        self.final_args = {}
        self.final_args.update(self.base_args)
        # also reset the bintemp validators so toggles between MPL-4 and 5+ don't cause validation errors
        params['galaxy']['bintemp'].validators = []

    def _get_url(self):
        ''' Retrieve the URL route from the map based on the request endpoint '''
        self._check_urlmap()
        blue, end = self.endpoint.split('.', 1)
        url = self.urlmap[blue][end]['url']
        # if the blueprint is not api, add/remove session option location
        if blue == 'api':
            if 'session' in parser.locations:
                pl = list(parser.locations)
                pl.remove('session')
                parser.locations = tuple(pl)
        else:
            if 'session' not in parser.locations:
                parser.locations += ('session', )

        return url

    def _extract_view_args(self):
        ''' Extract any view argument parameters contained in the URL route '''
        url = self._get_url()
        url_viewargs = re.findall(r'{(.*?)}', url)
        return url_viewargs

    def _add_param_args(self):
        ''' Adds all appropriate form arguments into dictionary for validation '''

        # get the url
        url = self._get_url()

        # check list or not
        self.use_params = [self.use_params] if not isinstance(self.use_params, (list, tuple)) else self.use_params

        for local_param in self.use_params:
            if local_param in params:
                # update param validation
                if self._required or self._setmissing:
                    newparams = self.update_param_validation(local_param)
                else:
                    newparams = deepcopy(params)
                # add to params final args
                self.final_args.update(newparams[local_param])

    def _set_params_required(self, subset):
        ''' Set the param validation required parameter '''

        # make list or not
        self._required = [self._required] if not isinstance(self._required, (list, tuple)) else self._required

        # update the required attribute
        for req_param in self._required:
            if req_param in subset.keys():
                subset[req_param].required = True
                subset[req_param].allow_none = False

            if req_param == 'bintemp':
                bintemps = self._get_bin_temps()
                subset[req_param].validate = validate.OneOf(bintemps)
                subset[req_param].validators.append(validate.OneOf(bintemps))

        return subset

    def _set_params_missing(self, subset):
        ''' Set the param validation missing parameter '''

        for miss_field in subset.values():
            miss_field.missing = None

        return subset

    def update_param_validation(self, name):
        ''' Update the validation of form params '''

        # deep copy the global parameter dict
        newparams = deepcopy(params)
        subset = newparams[name]

        # Set required params
        if self._required:
            subset = self._set_params_required(subset)

        # Set missing params
        if self._setmissing:
            subset = self._set_params_missing(subset)

        # return the new params
        newparams[name] = subset
        return newparams

    def _add_view_args(self):
        ''' Adds all appropriate View arguments into dictionary for validation '''
        local_viewargs = self._extract_view_args()

        # check if any local_view args need new validation
        props = ['bintype', 'template', 'property_name', 'channel']
        ismatch = set(local_viewargs) & set(props)
        if ismatch:
            self.update_view_validation()

        # add only the local view args to the final arg list
        if local_viewargs:
            for varg in local_viewargs:
                self.final_args.update({varg: viewargs[varg]})

    def _update_viewarg(self, name, choices):
        ''' Updates the global View arguments validator '''
        viewargs[name] = fields.String(required=True, location='view_args', validate=validate.OneOf(choices))
        # viewargs[name].validate = validate.OneOf(choices)

    def _get_bin_temps(self):
        ''' Gets the bintemps for a given release '''
        bintemps = dm[self.dapver].get_bintemps()
        return bintemps

    def update_view_validation(self):
        ''' Update the validation of DAP MPL specific names based on the datamodel '''

        # update all the dap datamodel specific options
        bintemps = self._get_bin_temps()
        bintypes = list(set([b.split('-', 1)[0] for b in bintemps]))
        temps = list(set([b.split('-', 1)[1] for b in bintemps]))
        properties = [gr.name for gr in dm[self.dapver].properties.extensions]
        channels = list(set([i.channel.name for i in dm[self.dapver].properties
                             if i.channel is not None])) + ['None']

        # update the global viewargs for each property
        propfields = {'bintype': bintypes, 'template': temps, 'property_name': properties,
                      'channel': channels}
        for key, val in propfields.items():
            self._update_viewarg(key, val)

    def create_args(self):
        ''' Build the final argument list for webargs validation '''

        # get the dapver
        #drpver, self.dapver = config.lookUpVersions(self.release)

        try:
            drpver, self.dapver = config.lookUpVersions(self.release)
        except MarvinError as e:
            pass

        # reset the final args
        self._reset_final_args()

        # add view args to the list
        self._add_view_args()

        # add form param args to the list
        if self.use_params:
            self._add_param_args()

    def _pop_kwargs(self, **kwargs):
        ''' Pop all non webargs kwargs out of the main kwarg dict '''

        webargs_kwargs = ['req', 'locations', 'as_kwargs', 'validate']
        tempkwargs = kwargs.copy()
        for key in kwargs:
            if key not in webargs_kwargs:
                tmp = tempkwargs.pop(key, None)
        self._main_kwargs = tempkwargs

    def _check_mainkwargs(self, **kwargs):
        self.use_params = kwargs.pop('use_params', None)
        self._required = kwargs.pop('required', None)
        self._setmissing = kwargs.pop('set_missing', None)

    def _get_release_endpoint(self, view):
        ''' get the release and endpoint if you can '''
        self.release = view._release
        self.endpoint = view._endpoint

    def check_args(self, **mainkwargs):
        ''' Checks the input view and parameter arguments for validation using webargs

        This is a decorator and modifies the standard webargs.flaskparser.use_args decorator.
        check_args by default will check and validation all View parameter arguments.  It will
        optionally check Request parameter arguments when the `use_params` keyword is set.

        Parameters:
            use_params (str|list):
                one (or list) of the param keys defining which sets of parameters to consider
                for validation
            required (str|list):
                string or list of string parameter names that are required as input

        Returns:
            A dictionary of arguments if all passed validation or an error dictionary specifying
            any missing parameters, or parameters that failed validation.  There errors are already
            caught properly in the Marvin system.

        Example:
            >>> # to validate all input view args
            >>> @av.check_args()
            >>>
            >>> # to validate input query request parameters
            >>> @av.check_args(use_params='query', required='searchfilter')

        '''
        # decorator used to grab the release and endpoint of the route
        def decorator(func):
            @functools.wraps(func)
            def wrapper(*args, **kwargs):
                # args and kwargs here are the view function args and kwargs
                self._get_release_endpoint(args[0])
                # check the kwargs for any parameters
                self._check_mainkwargs(**mainkwargs)
                # create the arguments dictionary
                self.create_args()
                # pop all the kwargs
                self._pop_kwargs(**mainkwargs)
                # pass into webargs use_args (use_args is a decorator in itself)
                newfunc = self.use_args(self.final_args, **self._main_kwargs)(func)
                return newfunc(*args, **kwargs)
            return wrapper

        return decorator

    def check_kwargs(self, **kwargs):
        kwargs['as_kwargs'] = True
        return self.check_args(**kwargs)

    def check_release(self, **kwargs):
        ''' Checks only the release '''
        return self.use_kwargs(self.base_args)

    def list_params(self, param_type=None):
        ''' List the globally defined parameters for validation
        '''
        total = {'viewargs': viewargs, 'params': params}

        if param_type == 'viewargs':
            return total['viewargs']
        elif param_type == 'params':
            return total['params']
        else:
            return total

    def manual_parse(self, view, req, **mainkwargs):
        ''' Manually parse the args

        Manually parses and validates the input arguements of a Flask route

        Parameters:
            view:
                The instance of the Flask View
            req:
                The Flask request object
            use_params (str|list):
                one (or list) of the param keys defining which sets of parameters to consider
                for validation
            required (str|list):
                string or list of string parameter names that are required as input
            makemulti (bool):
                If True, converts the arguments dictionary into an ImmutableMultiDict object

        Returns:
            A dictionary of arguments if all passed validation or an error dictionary specifying
            any missing parameters, or parameters that failed validation.  There errors are already
            caught properly in the Marvin system.

        Example:
            >>> # to validate on the Galaxy View page, and check the galaxy request parameters
            >>> # and require that the plateifu parameter is set
            >>> args = av.manual_parse(self, request, use_params='galaxy', required='plateifu')

        '''
        self._get_release_endpoint(view)
        url = self._get_url()
        self._check_mainkwargs(**mainkwargs)
        self.create_args()
        self._pop_kwargs(**mainkwargs)
        newargs = parser.parse(self.final_args, req, **self._main_kwargs)

        # see if we make it a multidict
        makemulti = mainkwargs.get('makemulti', None)
        if makemulti:
            from werkzeug.datastructures import ImmutableMultiDict
            newargs = ImmutableMultiDict(newargs.copy())
        return newargs


# Creates a context manager for executing code with DB off but then turning it on again after that.

@contextlib.contextmanager
def db_off():
    config.forceDbOff()
    yield
    config.forceDbOn()


def set_api_decorators():
    ''' Sets the API decorators '''
    if os.environ.get('PUBLIC_SERVER', None):
        from brain.utils.general.decorators import public
        decorators = [public]
    else:
        decorators = [fresh_jwt_required, validate_jwt]

    return decorators



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: Brian Cherinka, José Sánchez-Gallego, and Brett Andrews
# @Date: 2016-04-11
# @Filename: rss.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)
#
# @Last modified by: José Sánchez-Gallego (gallegoj@uw.edu)
# @Last modified time: 2018-07-30 19:42:17


from __future__ import division, print_function

import os
import warnings

import astropy.io.ascii
import astropy.table
import astropy.units
import astropy.wcs
import numpy
from astropy.io import fits

import marvin
from marvin.core.exceptions import MarvinError, MarvinUserWarning
from marvin.utils.datamodel.drp import datamodel_rss
from marvin.utils.datamodel.drp.base import Spectrum as SpectrumDataModel

from .core import MarvinToolsClass
from .cube import Cube
from .mixins import NSAMixIn
from .quantities.spectrum import Spectrum


class RSS(MarvinToolsClass, NSAMixIn, list):
    """A class to interface with a MaNGA DRP row-stacked spectra file.

    This class represents a fully reduced DRP row-stacked spectra object,
    initialised either from a file, a database, or remotely via the Marvin API.
    Instances of `.RSS` are a list of `.RSSFiber` objects, one for each fibre
    and exposure. `.RSSFiber` are initialised lazily, containing only basic
    information. They need to be initialised by calling `.RSSFiber.load`
    (unless `.RSS.autoload` is ``True``, in which case the instance is loaded
    when first accessed).

    In addition to the input arguments supported by `~.MarvinToolsClass` and
    `~.NSAMixIn`, this class accepts an ``autoload`` keyword argument that
    defines whether `.RSSFiber` objects should be automatically loaded when
    they are accessed.

    """

    _qualflag = 'DRP3QUAL'

    def __init__(self, input=None, filename=None, mangaid=None, plateifu=None,
                 mode=None, data=None, release=None, autoload=True,
                 drpall=None, download=None, nsa_source='auto'):

        MarvinToolsClass.__init__(self, input=input, filename=filename, mangaid=mangaid,
                                  plateifu=plateifu, mode=mode, data=data, release=release,
                                  drpall=drpall, download=download)

        NSAMixIn.__init__(self, nsa_source=nsa_source)

        #: An `astropy.table.Table` with the observing information associated
        #: with this RSS object.
        self.obsinfo = None

        #: If True, unloaded `.RSSFiber` instances are automatically loaded
        #: when accessed. Otherwise, they need to be loaded via `.RSSFiber.load`.
        self.autoload = autoload

        if self.data_origin == 'file':
            self._load_rss_from_file(data=self.data)
        elif self.data_origin == 'db':
            self._load_rss_from_db(data=self.data)
        elif self.data_origin == 'api':
            self._load_rss_from_api()

        Cube._init_attributes(self)

        # Checks that the drpver set in MarvinToolsClass matches the header
        header_drpver = self.header['VERSDRP3'].strip()
        header_drpver = 'v1_5_1' if header_drpver == 'v1_5_0' else header_drpver
        assert header_drpver == self._drpver, ('mismatch between cube._drpver={0} '
                                               'and header drpver={1}'.format(self._drpver,
                                                                              header_drpver))

        # EXPNUM in obsinfo is a string. Cast it to int
        self.obsinfo['EXPNUM'] = self.obsinfo['EXPNUM'].astype(numpy.int32)

        # Inits self as an empty list.
        list.__init__(self, [])
        self._populate_fibres()

    def _set_datamodel(self):
        """Sets the datamodel for DRP."""

        self.datamodel = datamodel_rss[self.release.upper()]
        self._bitmasks = datamodel_rss[self.release.upper()].bitmasks

    def __repr__(self):
        """Representation for RSS."""

        return ('<Marvin RSS (mangaid={self.mangaid!r}, plateifu={self.plateifu!r}, '
                'mode={self.mode!r}, data_origin={self.data_origin!r})>'.format(self=self))

    def __getitem__(self, fiberid):
        """Returns the `.RSSFiber` whose fiberid matches the input."""

        rssfiber = super(RSS, self).__getitem__(fiberid)

        if self.autoload and not rssfiber.loaded:
            rssfiber.load()

        return rssfiber

    def _getFullPath(self):
        """Returns the full path of the file in the tree."""

        if not self.plateifu:
            return None

        plate, ifu = self.plateifu.split('-')

        return super(RSS, self)._getFullPath('mangarss', ifu=ifu, drpver=self._drpver, plate=plate, wave='LOG')

    def download(self):
        """Downloads the cube using sdss_access - Rsync"""

        if not self.plateifu:
            return None

        plate, ifu = self.plateifu.split('-')

        return super(RSS, self).download('mangarss', ifu=ifu, drpver=self._drpver, plate=plate, wave='LOG')

    def getCube(self):
        """Returns the `~marvin.tools.cube.Cube` associated with this RSS."""

        return Cube(plateifu=self.plateifu, mode=self.mode, release=self.release)

    def load_all(self):
        """Loads all the `.RSSFiber` associated to this `.RSS` instance."""

        for rssfiber in self:
            if not rssfiber.loaded:
                rssfiber.load()

    def select_fibers(self, exposure_no=None, set=None, mjd=None):
        """Selects fibres that match one or multiple of the input parameters.

        Parameters
        ----------
        exposure_no : int
            The exposure number. Ignored if ``None``.
        set : int
            The set id of the exposure. Ignored if ``None``.
        mjd : int
            The MJD of the exposure. Ignored if ``None``.

        Returns
        -------
        rssfibers : list
            A list of `.RSSFiber` instances whose obsinfo matches all the input
            parameters. The `.RSS.autoload` option is respected.

        Example
        -------

            >>> rss = marvin.tools.RSS('8485-1901')
            >>> fibers = rss.select_fibers(set=2)
            >>> fibers
            [<RSSFiber [ 2.22306705, 11.84955406,  9.65761662, ...,  0.        ,
                         0.        ,  0.        ] 1e-17 erg / (Angstrom cm2 fiber s)>,
            <RSSFiber [2.18669987, 1.4861778 , 2.55065155, ..., 0.        , 0.        ,
                       0.        ] 1e-17 erg / (Angstrom cm2 fiber s)>,
            <RSSFiber [2.75228763, 5.53485441, 2.31695175, ..., 0.        , 0.        ,
                       0.        ] 1e-17 erg / (Angstrom cm2 fiber s)>]

        """

        mask_exp = (self.obsinfo['EXPNUM'].astype(int) == exposure_no) if exposure_no else True
        mask_set = (self.obsinfo['SET'].astype(int) == set) if set else True
        mask_mjd = (self.obsinfo['MJD'].astype(int) == mjd) if mjd else True

        mask = mask_exp & mask_set & mask_mjd
        valid_exposures = numpy.where(mask)[0]

        n_exposures = len(self.obsinfo)
        n_fibres_per_exposure = self._nfibers // n_exposures
        fibre_to_exposure = numpy.arange(self._nfibers) // n_fibres_per_exposure

        fibres_in_valid_exposures = numpy.where(numpy.in1d(fibre_to_exposure, valid_exposures))[0]

        return [self[ii] for ii in fibres_in_valid_exposures]

    def _load_rss_from_file(self, data=None):
        """Initialises the RSS object from a file."""

        if data is not None:
            assert isinstance(data, fits.HDUList), 'data is not an HDUList object'
        else:
            try:
                self.data = fits.open(self.filename)
            except (IOError, OSError) as err:
                raise OSError('filename {0} cannot be found: {1}'.format(self.filename, err))

        self.header = self.data[1].header
        self.wcs = astropy.wcs.WCS(self.header)
        self.wcs = self.wcs.dropaxis(1)  # The header creates an empty axis for the exposures.

        # Confirm that this is a RSS file
        assert 'XPOS' in self.data and self.header['CTYPE1'] == 'WAVE-LOG', \
            'invalid file type. It does not appear to be a LOGRSS.'

        self._wavelength = self.data['WAVE'].data
        self._shape = None
        self._nfibers = self.data['FLUX'].shape[0]

        self.obsinfo = astropy.table.Table(self.data['OBSINFO'].data)

        Cube._do_file_checks(self)

    def _load_rss_from_db(self, data=None):
        """Initialises the RSS object from the DB.

        At this time the DB does not contain enough information to successfully
        instantiate a RSS object so we hack the data access mode to try to use
        files. For users this should be irrelevant since they rarely will have
        a Marvin DB. For the API, it means the access to RSS data will happen
        via files.

        """

        warnings.warn('DB mode is not working for RSS. Trying file access mode.',
                      MarvinUserWarning)

        fullpath = self._getFullPath()
        if fullpath and os.path.exists(fullpath):
            self.filename = fullpath
            self.data_origin = 'file'
            self._load_rss_from_file()
        else:
            raise MarvinError('cannot find a valid RSS file for '
                              'plateifu={self.plateifu!r}, release={self.release!r}'
                              .format(self=self))

    def _load_rss_from_api(self):
        """Initialises the RSS object using the remote API."""

        # Checks that the RSS exists.
        routeparams = {'name': self.plateifu}
        url = marvin.config.urlmap['api']['getRSS']['url'].format(**routeparams)

        try:
            response = self._toolInteraction(url.format(name=self.plateifu))
        except Exception as ee:
            raise MarvinError('found a problem when checking if remote RSS '
                              'exists: {0}'.format(str(ee)))

        data = response.getData()

        self.header = fits.Header.fromstring(data['header'])
        self.wcs = astropy.wcs.WCS(fits.Header.fromstring(data['wcs_header']))
        self._wavelength = data['wavelength']
        self._nfibers = data['nfibers']

        self.obsinfo = astropy.io.ascii.read(data['obsinfo'])

        if self.plateifu != data['plateifu']:
            raise MarvinError('remote RSS has a different plateifu!')

        return

    def _populate_fibres(self):
        """Populates the internal list of fibres."""

        n_exposures = len(self.obsinfo)
        n_fibres_per_exposure = self._nfibers // n_exposures

        for fiberid in range(self._nfibers):
            exp_index = fiberid // n_fibres_per_exposure
            exp_obsinfo = self.obsinfo[[exp_index]]
            self.append(RSSFiber(fiberid, self, self._wavelength, load=False,
                                 obsinfo=exp_obsinfo, pixmask_flag=self.header['MASKNAME']))


class RSSFiber(Spectrum):
    """A `~astropy.units.Quantity` representing a fibre observation.

    Represents the spectral flux observed though a fibre, and associated with
    an `.RSS` object. In addition to the flux, it contains information about
    the inverse variance, mask, and other associated spectra defined in the
    datamodel.

    Parameters
    ----------
    fiberid : int
        The fiberid (0-indexed row in the parent `.RSS` object) for this fibre
        observation.
    rss : `.RSS`
        The parent `.RSS` object with which this fibre observation is
        associated.
    wavelength : numpy.ndarray
        The wavelength positions of each array element, in Angstrom.
    load : bool
        Whether the information in the `.RSSFiber` should be loaded during
        instantiation. Defaults to lazy loading (use `.RSSFiber.load` to
        load the fibre information).
    obsinfo : astropy.table.Table
        A `~astropy.table.Table` with the information for the exposure to
        which this fibre observation belongs.
    kwargs : dict
        Additional keyword arguments to be passed to `.Spectrum`.

    """

    def __new__(cls, fiberid, rss, wavelength, pixmask_flag=None, load=False,
                obsinfo=None, **kwargs):

        # For now we instantiate a mostly empty Spectrum. Proper instantiation
        # will happen in load().

        array_size = len(wavelength)

        obj = super(RSSFiber, cls).__new__(
            cls, numpy.zeros(array_size, dtype=numpy.float64), wavelength,
            scale=None, unit=None,)

        obj._extra_attributes = ['fiberid', 'rss', 'loaded', 'obsinfo']
        obj._spectra = []

        return obj

    def __init__(self, fiberid, rss, wavelength, pixmask_flag=None, load=False,
                 obsinfo=None, **kwargs):

        self.fiberid = fiberid
        self.rss = rss
        self.obsinfo = obsinfo

        self.pixmask_flag = pixmask_flag

        self.loaded = False
        if load:
            self.load()

    def __repr__(self):

        if not self.loaded:
            return ('<RSSFiber (plateifu={self.rss.plateifu!r}, '
                    'fiberid={self.fiberid!r}, loaded={self.loaded!r})>'.format(self=self))
        else:
            return super(RSSFiber, self).__repr__()

    def __array_finalize__(self, obj):

        if obj is None:
            return

        super(RSSFiber, self).__array_finalize__(obj)

        # Adds _extra_attributes from the previous object.
        if hasattr(obj, '_extra_attributes'):
            for attr in obj._extra_attributes:
                setattr(self, attr, getattr(obj, attr, None))
        self._extra_attributes = getattr(obj, '_extra_attributes', None)

        # Adds the additional spectra from the previous object.
        if hasattr(obj, '_spectra'):
            for spectrum in obj._spectra:
                setattr(self, spectrum, getattr(obj, spectrum, None))
        self._spectra = getattr(obj, '_spectra', None)

    def __getitem__(self, sl):

        new_obj = super(RSSFiber, self).__getitem__(sl)

        for spectra_name in self._spectra:
            current_spectrum = getattr(self, spectra_name, None)
            new_spectrum = None if current_spectrum is None else current_spectrum.__getitem__(sl)
            setattr(new_obj, spectra_name, new_spectrum)

        return new_obj

    def load(self):
        """Loads the fibre information."""

        assert self.loaded is False, 'object already loaded.'

        # Depending on whether the parent RSS is a file or API-populated, we
        # select the data to use.
        if self.rss.data_origin == 'file':

            # If the data origin is a file we use the HDUList in rss.data
            rss_data = self.rss.data

        elif self.rss.data_origin == 'api':

            # If data origin is the API, we make a request for the data
            # associated with this fiberid for all the extensions in the file.

            url = marvin.config.urlmap['api']['getRSSFiber']['url']

            try:
                response = self.rss._toolInteraction(url.format(name=self.rss.plateifu,
                                                                fiberid=self.fiberid))
            except Exception as ee:
                raise MarvinError('found a problem retrieving RSS fibre data for '
                                  'plateifu={!r}, fiberid={!r}: {}'.format(
                                      self.rss.plateifu, self.fiberid, str(ee)))

            api_data = response.getData()

            # Create a quick and dirty HDUList from the API data so that we
            # can parse it in the same way as if the data origin is file.
            rss_data = astropy.io.fits.HDUList([astropy.io.fits.PrimaryHDU()])
            for ext in api_data:
                rss_data.append(astropy.io.fits.ImageHDU(data=api_data[ext], name=ext.upper()))

        else:
            raise ValueError('invalid data_origin={!r}'.format(self.rss.data_origin))

        # Compile a list of all RSS datamodel extensions, either RSS or spectra
        datamodel_extensions = self.rss.datamodel.rss + self.rss.datamodel.spectra

        for extension in datamodel_extensions:

            # Retrieve the value (and mask and ivar, if associated) for each extension.
            value, ivar, mask = self._get_extension_data(extension, rss_data,
                                                         data_origin=self.rss.data_origin)

            if extension.name == 'flux':

                self.value[:] = value[:]
                self.ivar = ivar
                self.mask = mask
                self._set_unit(extension.unit)

            else:

                new_spectrum = Spectrum(value, self.wavelength, ivar=ivar, mask=mask,
                                        unit=extension.unit)
                setattr(self, extension.name, new_spectrum)

                self._spectra.append(extension.name)

        self.loaded = True

    def _get_extension_data(self, extension, data, data_origin='file'):
        """Returns the value of an extension for this fibre, either from file or API.

        Parameters
        ----------
        extension : datamodel object
            The datamodel object containing the information for the extension
            we want to retrieve.
        data : ~astropy.io.fits.HDUList
            An `~astropy.io.fits.HDUList` object containing the RSS
            information.

        """

        # Determine if this is an RSS datamodel object or an spectrum.
        # If the origin is the API, the extension data contains a single spectrum,
        # not a row-stacked array, so we consider it a 1D array.
        is_extension_data_1D = isinstance(extension, SpectrumDataModel) or data_origin == 'api'

        value = data[extension.fits_extension()].data

        if extension.has_mask():
            mask = data[extension.fits_extension('mask')].data
        else:
            mask = None

        if hasattr(extension, 'has_ivar') and extension.has_ivar():
            ivar = data[extension.fits_extension('ivar')].data
        elif hasattr(extension, 'has_std') and extension.has_std():
            std = data[extension.fits_extension('std')].data
            ivar = 1. / (std**2)
        else:
            ivar = None

        # If this is an RSS, gets the right row in the stacked spectra.
        if not is_extension_data_1D:
            value = value[self.fiberid, :]
            mask = mask[self.fiberid, :] if mask is not None else None
            ivar = ivar[self.fiberid, :] if ivar is not None else None

        return value, ivar, mask

    @property
    def masked(self):
        """Return a masked array where the mask is greater than zero."""

        assert self.mask is not None, 'mask is not set'

        return numpy.ma.array(self.value, mask=(self.mask > 0))

    def descale(self):
        """Returns a copy of the object in which the scale is unity.

        Note that this only affects to the core value of this quantity.
        Associated array attributes will not be modified.

        Example:

            >>> fiber.unit
            Unit("1e-17 erg / (Angstrom cm2 fiber s)")
            >>> fiber[100]
            <RSSFiber 0.270078063011169 1e-17 erg / (Angstrom cm2 fiber s)>
            >>> fiber_descaled = fiber.descale()
            >>> fiber_descaled.unit
            Unit("Angstrom cm2 fiber s")
            >>> fiber[100]
            <RSSFiber 2.70078063011169e-18 erg / (Angstrom cm2 fiber s)>

        """

        if self.unit.scale == 1:
            return self

        value_descaled = self.value * self.unit.scale
        value_unit = astropy.units.CompositeUnit(1, self.unit.bases, self.unit.powers)

        if self.ivar is not None:
            ivar_descaled = self.ivar / (self.unit.scale ** 2)
        else:
            ivar_descaled = None

        copy_of_self = self.copy()
        copy_of_self.value[:] = value_descaled
        copy_of_self.ivar = ivar_descaled
        copy_of_self._set_unit(value_unit)

        return copy_of_self



# !/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Filename: MPL8.py
# Project: dap
# Author: Brian Cherinka
# Created: Friday, 8th February 2019 2:10:30 pm
# License: BSD 3-clause "New" or "Revised" License
# Copyright (c) 2019 Brian Cherinka
# Last Modified: Friday, 17th May 2019 2:42:07 pm
# Modified By: Brian Cherinka


from __future__ import print_function, division, absolute_import

import copy
from astropy import units as u
from marvin.utils.datamodel.maskbit import get_maskbits
from .base import DAPDataModel, Template, spaxel as spaxel_unit
from .base import Model, Channel, MultiChannelProperty
from .MPL6 import SPX, HYB10, VOR10, MPL6_maps, binid_properties
from .MPL6 import MPL6_emline_channels, oii_channel, oiid_channel

# update Template for MPL-8
MILESHC_MILESHC = Template('MILESHC-MILESHC',
                           description=('Set of templates derived from the MILES library, '
                                        'used for both the stellar kinematics and emission lines.'))


# new figure of merit channels
fom_channels = [
    Channel('rms', formats={'string': 'RMS of residuals for fitted pixels',
                            'latex': r'RMS of residuals'},
            description='RMS of the residuals for all fitted pixels in the stellar continuum.',
            unit=u.erg / u.s / (u.cm ** 2) / u.Angstrom / spaxel_unit, scale=1e-17, idx=0),
    Channel('frms', formats={'string': 'RMS of the fractional residuals',
                             'latex': r'RMS of the fractional residuals'},
            description='Fractional residuals for the stellar continuum fit.', idx=1),
    Channel('rchi2', formats={'string': 'Stellar continuum reduced chi-square',
                              'latex': r'Stellar\ continuum\ reduced\ \chi^2'},
            description='Reduced chi-square of the stellar continuum fit.', idx=2),
    Channel('fresid_68th_percentile', db_name='68th_perc_frac_resid',
            formats={'string': '68th percentile',
                     'latex': r'68^{th} percentile'},
            description='68%% growth of the fractional residuals between the model and data.', idx=3),
    Channel('fresid_99th_percentile', db_name='99th_perc_frac_resid',
            formats={'string': '99th percentile Frac. Residuals',
                     'latex': r'99^{th} percentile Frac. Residuals'},
            description='99%% growth of the fractional residuals between the model and data.', idx=4),
    Channel('fresid_max', db_name='max_frac_resid',
            formats={'string': 'Max Fractional Residual',
                     'latex': r'Max Fractional Residual'},
            description='Maximum growth of the fractional residuals between the model and data.', idx=5),
    Channel('per_pix_chi_68th_percentile', db_name='68th_perc_per_pix_chi',
            formats={'string': '68th percentile',
                     'latex': r'68^{th} percentile'},
            description='68%% growth of the error-normalized residuals', idx=6),
    Channel('per_pix_chi_99th_percentile', db_name='99th_perc_per_pix_chi',
            formats={'string': '99th percentile',
                     'latex': r'99^{th} percentile'},
            description='99%% growth of the error-normalized residuals', idx=7),
    Channel('per_pix_chi_max', db_name='max_per_pix_chi',
            formats={'string': 'Max Error-Normalized Residuals',
                     'latex': r'Max Error-Normalized Residuals'},
            description='Maximum growth of the error-normalized residuals', idx=8),
]

# new emission line channels
new_channels = [
    Channel('hei_3889', formats={'string': 'HeI 3889',
                                 'latex': r'$\forb{He\,I}\;\lambda 3889$'}, idx=22),
    Channel('ni_5199', formats={'string': 'NI 5199',
                                'latex': r'$\forb{N\,I}\;\lambda 5199$'}, idx=23),
    Channel('ni_5201', formats={'string': 'NI 5201',
                                'latex': r'$\forb{N\,I}\;\lambda 5201$'}, idx=24),
]

MPL8_emline_channels = copy.deepcopy(MPL6_emline_channels) + new_channels

new_properties = [
    # New Emission Line extensions
    MultiChannelProperty('emline_ga', ivar=False, mask=False,
                         channels=[oii_channel] + MPL8_emline_channels,
                         formats={'string': 'Amplitude of Fitted Gaussians'},
                         unit=u.erg / u.s / (u.cm ** 2) / u.Angstrom / spaxel_unit, scale=1e-17,
                         binid=binid_properties[3],
                         description='Amplitude of the fitted Gaussian emission lines.'),
    MultiChannelProperty('emline_ganr', ivar=False, mask=False,
                         channels=[oii_channel] + MPL8_emline_channels,
                         formats={'string': 'Amplitude Over Noise of Fitted Gaussians'},
                         binid=binid_properties[3],
                         description='Amplitude Over Noise of the fitted Gaussian emission lines.'),
    MultiChannelProperty('emline_lfom', ivar=False, mask=False,
                         channels=[oii_channel] + MPL8_emline_channels,
                         formats={'string': 'Emission Line Reduced Chi-Square'},
                         binid=binid_properties[3],
                         description='Reduced chi-square in 15 pixel windows around each fitted emission line'),
    # Figure of Merit Extensions in MAPS
    MultiChannelProperty('emline_fom', ivar=False, mask=False,
                         channels=fom_channels,
                         formats={'string': 'Emission Line FOM'},
                         unit=u.dimensionless_unscaled,
                         description='Full spectrum figures-of-merit for emission line fitting'),
    MultiChannelProperty('stellar_fom', ivar=False, mask=False,
                         channels=fom_channels, formats={
                             'string': 'Stellar FOM'},
                         unit=u.dimensionless_unscaled,
                         description='Full spectrum figures-of-merit for stellar continuum fitting'),
    # New Stellar SigmaCorr
    MultiChannelProperty('stellar_sigmacorr', ivar=False, mask=False,
                         channels=[Channel('resolution_difference',
                                           formats={'string': 'Resolution Difference'}, idx=0),
                                   Channel('fit',
                                           formats={'string': 'Fit'}, idx=1)],
                         unit=u.km / u.s,
                         formats={'string': 'Stellar sigma correction',
                                  'latex': r'Stellar $\sigma$ correction'},
                         description='Quadrature correction for STELLAR_SIGMA to obtain the '
                                     'astrophysical velocity dispersion.'),
]

emline_props = [
    MultiChannelProperty('emline_sflux', ivar=True, mask=True,
                         channels=[oiid_channel] + MPL8_emline_channels,
                         formats={'string': 'Emission line summed flux'},
                         unit=u.erg / u.s / (u.cm ** 2) / spaxel_unit, scale=1e-17,
                         binid=binid_properties[3],
                         description='Non-parametric summed flux for emission lines.'),
    MultiChannelProperty('emline_sew', ivar=True, mask=True,
                         channels=[oiid_channel] + MPL8_emline_channels,
                         formats={'string': 'Emission line EW'},
                         unit=u.Angstrom,
                         binid=binid_properties[3],
                         description='Emission line non-parametric equivalent '
                                     'widths measurements.'),
    MultiChannelProperty('emline_gflux', ivar=True, mask=True,
                         channels=[oii_channel] + MPL8_emline_channels,
                         formats={'string': 'Emission line Gaussian flux'},
                         unit=u.erg / u.s / (u.cm ** 2) / spaxel_unit, scale=1e-17,
                         binid=binid_properties[3],
                         description='Gaussian profile integrated flux for emission lines.'),
    MultiChannelProperty('emline_gvel', ivar=True, mask=True,
                         channels=[oii_channel] + MPL8_emline_channels,
                         formats={'string': 'Emission line Gaussian velocity'},
                         unit=u.km / u.s,
                         binid=binid_properties[3],
                         description='Gaussian profile velocity for emission lines.'),
    MultiChannelProperty('emline_gew', ivar=True, mask=True,
                         channels=[oii_channel] + MPL8_emline_channels,
                         formats={'string': 'Emission line Gaussian EW'},
                         unit=u.Angstrom,
                         binid=binid_properties[3],
                         description='Gaussian-fitted equivalent widths measurements '
                                     '(based on EMLINE_GFLUX).'),
    MultiChannelProperty('emline_gsigma', ivar=True, mask=True,
                         channels=[oii_channel] + MPL8_emline_channels,
                         formats={'string': 'Emission line Gaussian sigma',
                                  'latex': r'Emission line Gaussian $\sigma$'},
                         unit=u.km / u.s,
                         binid=binid_properties[3],
                         description='Gaussian profile velocity dispersion for emission lines; '
                                     'must be corrected using EMLINE_INSTSIGMA.'),
    MultiChannelProperty('emline_instsigma', ivar=False, mask=False,
                         channels=[oii_channel] + MPL8_emline_channels,
                         formats={'string': 'Emission line instrumental sigma',
                                  'latex': r'Emission line instrumental $\sigma$'},
                         unit=u.km / u.s,
                         binid=binid_properties[3],
                         description='Instrumental dispersion at the fitted line center.'),
    MultiChannelProperty('emline_tplsigma', ivar=False, mask=False,
                         channels=[oii_channel] + MPL8_emline_channels,
                         formats={'string': 'Emission line template instrumental sigma',
                                  'latex': r'Emission line template instrumental $\sigma$'},
                         unit=u.km / u.s,
                         binid=binid_properties[3],
                         description='The dispersion of each emission line used in '
                                     'the template spectra'),
]

# Update Maps Properties
remove = ['stellar_cont_rchi2', 'stellar_sigmacorr', 'stellar_cont_fresid']
MPL8_maps = [m for m in MPL6_maps if m.name not in remove and 'emline' not in m.name]
MPL8_maps += emline_props + new_properties

# Update Model Properties
MPL8_models = [
    Model('binned_flux', 'FLUX', 'WAVE', extension_ivar='IVAR',
          extension_mask='MASK', unit=u.erg / u.s / (u.cm ** 2) / u.Angstrom / spaxel_unit,
          scale=1e-17, formats={'string': 'Binned flux'},
          description='Flux of the binned spectra',
          binid=binid_properties[0]),
    Model('full_fit', 'MODEL', 'WAVE', extension_ivar=None,
          extension_mask='MODEL_MASK', unit=u.erg / u.s / (u.cm ** 2) / u.Angstrom / spaxel_unit,
          scale=1e-17, formats={'string': 'Best fitting model'},
          description='The best fitting model spectra (sum of the fitted '
                      'continuum and emission-line models)',
          binid=binid_properties[0]),
    Model('emline_fit', 'EMLINE', 'WAVE', extension_ivar=None,
          extension_mask=None,
          unit=u.erg / u.s / (u.cm ** 2) / u.Angstrom / spaxel_unit,
          scale=1e-17, formats={'string': 'Emission line model spectrum'},
          description='The model spectrum with only the emission lines.',
          binid=binid_properties[3]),
    Model('stellar_fit', 'STELLAR', 'WAVE', extension_ivar=None,
          extension_mask='STELLAR_MASK',
          unit=u.erg / u.s / (u.cm ** 2) / u.Angstrom / spaxel_unit,
          scale=1e-17, formats={'string': 'Stellar continuum model spectrum'},
          description='The model spectrum with only the stellar continuum',
          binid=binid_properties[1])
]

# MPL-8 DapDataModel goes here
MPL8 = DAPDataModel('2.3.0', aliases=['MPL-8', 'MPL8'],
                    bintypes=[SPX, HYB10, VOR10],
                    db_only=[HYB10],
                    templates=[MILESHC_MILESHC],
                    properties=MPL8_maps,
                    models=MPL8_models,
                    bitmasks=get_maskbits('MPL-8'),
                    default_bintype='HYB10',
                    default_template='MILESHC-MILESHC',
                    property_table='SpaxelProp8',
                    default_binid=copy.deepcopy(binid_properties[0]),
                    default_mapmask=['NOCOV', 'UNRELIABLE', 'DONOTUSE'],
                    qual_flag='DAPQUAL')



# !usr/bin/env python2
# -*- coding: utf-8 -*-
#
# Licensed under a 3-clause BSD license.
#
# @Author: Brian Cherinka
# @Date:   2017-01-22 20:17:33
# @Last modified by:   Brian Cherinka
# @Last Modified time: 2018-07-16 11:52:02

from __future__ import print_function, division, absolute_import
import requests
import os
try:
    from urlparse import urljoin
except ImportError:
    from urllib.parse import urljoin
try:
    import matplotlib.pyplot as plt
    import mpl_toolkits.axes_grid1
    pyplot = True
except ImportError:
    pyplot = False
try:
    import ipwhois
except ImportError as e:
    ipwhois = None


class Sentry(object):
    ''' Interface to the Sentry API to grab and plot info

    Allows to query the Marvin2 Sentry page using their API.
    Can retrieve info on projects, events, issues, tags, etc.

    See doc here:
    https://docs.sentry.io/api/

    Parameters:
        org (str):
            Sentry organization name
        project (str):
            Sentry project name
        authtoken (str):
            token string for Sentry web api authentication

    '''

    def __init__(self, org='manga', project='marvin', authtoken=None):
        self.org = org
        self.project = project
        self.authtoken = authtoken
        self.data = None
        self.users = None
        self.tagvals = None
        self.ips = None
        self.locations = None
        self.response = None
        self.base_url = 'https://sentry.io/api/0/'
        self.header = {'Authorization': 'Bearer {authtoken}'}

    def check_auth(self):
        ''' checks and sets the authentication '''
        if self.authtoken:
            self.header['Authorization'] = self.header['Authorization'].format(**{'authtoken': self.authtoken})
        else:
            raise NameError('No authtoken set.  Please get one from the Sentry web api and set it!')

    def send_request(self, url, params=None):
        ''' Send the API request

        Parameters:
            url (str):
                The url to submit a request to
            params (list):
                list of optional parameters
        '''
        self.check_auth()
        self.response = requests.get(url, headers=self.header, params=params)
        self.check_response()

    def check_response(self):
        ''' checks the response

        Raises an error if the status is not 200. Otherwise sets gets the json data

        '''
        if not self.response.raise_for_status():
            self.data = self.response.json()

    def get_events(self):
        ''' Get the events for a projet '''
        url = urljoin(self.base_url, os.path.join('projects', self.org, self.project, 'events/'))
        self.send_request(url)
        self.events = self.data

    def get_users(self):
        ''' Get the users for a projet '''
        url = urljoin(self.base_url, os.path.join('projects', self.org, self.project, 'users/'))
        self.send_request(url)
        self.users = self.data

    def get_ips(self):
        ''' Get the ip addresses of a list of users '''
        if not self.users:
            self.get_users()
        self.ips = [u['ipAddress'] for u in self.users if u['ipAddress']]

    def lookup_ips(self, ip=None, method='whois'):
        ''' Look up the locations of the ips '''

        self.locations = []

        if ip and ip != '127.0.0.1':
            self.locations.append(self.get_ip_dict(ip, method=method))
        else:
            for ip in self.ips:
                if ip is not None and ip != '127.0.0.1':
                    self.locations.append(self.get_ip_dict(ip, method=method))

    def get_ip_dict(self, ip, method='whois'):
        ''' Get the ip lookup dictionary '''

        if not ipwhois:
            raise ImportError('Cannot look up ips.  You do not have the ipwhois package installed!')

        assert method in ['whois', 'rdap'], 'Method must either be rdap or whois'

        ipwho = ipwhois.IPWhois(ip)
        self.ipmethod = method
        ipdict = None
        if method == 'whois':
            try:
                ipdict = ipwho.lookup_whois()
            except Exception as e:
                print('Could not lookup ip {0}: {1}'.format(ip, e))
        elif method == 'rdap':
            try:
                ipdict = ipwho.lookup_rdap()
            except Exception as e:
                print('Could not lookup ip {0}: {1}'.format(ip, e))
        return ipdict

    def extract_locations(self):
        ''' Extraction the location info from the ip output '''
        self.places = []
        for loc in self.locations:
            nets = loc['nets'][0] if len(loc['nets']) > 0 else None
            locdict = {'asn_country_code': loc['asn_country_code'],
                       'place': {'city': nets['city'] if nets else None,
                                 'state': nets['state'] if nets else None,
                                 'country': nets['country'] if nets else None}}
            self.places.append(locdict)

    def get_project_tags(self):
        ''' Get the tags for a project '''
        url = urljoin(self.base_url, os.path.join('projects', self.org, self.project, 'tags/'))
        self.send_request(url)
        self.tags = self.data

    def list_tags(self):
        ''' List the tags available '''
        if self.tags:
            print([t['key'] for t in self.tags])
        else:
            print('No tags available')

    def get_tag_values(self, key=None):
        ''' Get the values for a given tag

        Parameters:
            key (str):
                The tag key to request the values for.  If not set, loops over all tags
        '''
        self.tagvals = {}

        if key is not None:
            self.tagvals[key] = self.get_value(key)
        else:
            for tag in self.tags:
                key = tag['key']
                self.tagvals[key] = self.get_value(key)

    def get_value(self, key):
        ''' Gets the values for single tag key

        Parameters:
            key (str):
                the tag key

        Returns:
            list of dictionaries of all values given the tag key

        '''
        url = os.path.join(self.base_url, 'projects', self.org, self.project, 'tags', key, 'values/')
        self.send_request(url)
        return self.data

    def get_place(self, place):
        ''' get a place from the places dictionary '''

        assert place in ['country', 'city', 'state'], 'Place must be either country, city, or state'
        return [p['place'][place] for p in self.places if p['place'][place]]

    def count_data(self, data):
        ''' Count the generic data and return a list of tuples '''
        return {(x, data.count(x)) for x in data}

    def get_pie_data(self, data):
        ''' Return the labels and sizes needed for the pie chart '''
        labels = [v[0] for v in data]
        sizes = [v[1] for v in data]
        return labels, sizes

    def plot_pie(self, key, data=None):
        ''' Makes a matplotlib pie chart for a given tag key '''
        if self.tagvals and key in self.tagvals:
            vals = self.tagvals[key]
            labels = [v['name'] for v in vals]
            sizes = [v['count'] for v in vals]
        elif key == 'place':
            assert data is not None, 'Must provide some data to extract'
            vals = self.count_data(data)
            labels, sizes = self.get_pie_data(vals)
        elif key == 'baseurl':
            vals = self.tagvals['url']
            names = [v['name'] for v in vals]
            labels = set(['sas' if 'sas.sdss.org' in n else 'api' for n in names])
            stats = [{'type': 'web' if l == 'sas' else l, 'count': sum(d['count'] for d in vals if l in d['name'])} for l in set(labels)]
            labels = [s['type'] for s in stats]
            sizes = [s['count'] for s in stats]
        else:
            sizes = None
            raise NameError('No sizes or labels found.  Make sure the data is somewhere')

        plt.clf()
        plt.pie(sizes, labels=labels, shadow=True, autopct='%1.1f%%', startangle=90)
        plt.axis('equal')
        plt.tight_layout()
        plt.show()




